# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Behdad Esfahbod

from fontTools.misc.textTools import bytesjoin, safeEval
from . import DefaultTable
import array
from collections import namedtuple
import struct
import sys


class table_C_P_A_L_(DefaultTable.DefaultTable):
    """Color Palette table

    The ``CPAL`` table contains a set of one or more color palettes. The color
    records in each palette can be referenced by the ``COLR`` table to specify
    the colors used in a color glyph.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/cpal
    """

    NO_NAME_ID = 0xFFFF
    DEFAULT_PALETTE_TYPE = 0

    def __init__(self, tag=None):
        DefaultTable.DefaultTable.__init__(self, tag)
        self.palettes = []
        self.paletteTypes = []
        self.paletteLabels = []
        self.paletteEntryLabels = []

    def decompile(self, data, ttFont):
        (
            self.version,
            self.numPaletteEntries,
            numPalettes,
            numColorRecords,
            goffsetFirstColorRecord,
        ) = struct.unpack(">HHHHL", data[:12])
        assert (
            self.version <= 1
        ), "Version of CPAL table is higher than I know how to handle"
        self.palettes = []
        pos = 12
        for i in range(numPalettes):
            startIndex = struct.unpack(">H", data[pos : pos + 2])[0]
            assert startIndex + self.numPaletteEntries <= numColorRecords
            pos += 2
            palette = []
            ppos = goffsetFirstColorRecord + startIndex * 4
            for j in range(self.numPaletteEntries):
                palette.append(Color(*struct.unpack(">BBBB", data[ppos : ppos + 4])))
                ppos += 4
            self.palettes.append(palette)
        if self.version == 0:
            offsetToPaletteTypeArray = 0
            offsetToPaletteLabelArray = 0
            offsetToPaletteEntryLabelArray = 0
        else:
            pos = 12 + numPalettes * 2
            (
                offsetToPaletteTypeArray,
                offsetToPaletteLabelArray,
                offsetToPaletteEntryLabelArray,
            ) = struct.unpack(">LLL", data[pos : pos + 12])
        self.paletteTypes = self._decompileUInt32Array(
            data,
            offsetToPaletteTypeArray,
            numPalettes,
            default=self.DEFAULT_PALETTE_TYPE,
        )
        self.paletteLabels = self._decompileUInt16Array(
            data, offsetToPaletteLabelArray, numPalettes, default=self.NO_NAME_ID
        )
        self.paletteEntryLabels = self._decompileUInt16Array(
            data,
            offsetToPaletteEntryLabelArray,
            self.numPaletteEntries,
            default=self.NO_NAME_ID,
        )

    def _decompileUInt16Array(self, data, offset, numElements, default=0):
        if offset == 0:
            return [default] * numElements
        result = array.array("H", data[offset : offset + 2 * numElements])
        if sys.byteorder != "big":
            result.byteswap()
        assert len(result) == numElements, result
        return result.tolist()

    def _decompileUInt32Array(self, data, offset, numElements, default=0):
        if offset == 0:
            return [default] * numElements
        result = array.array("I", data[offset : offset + 4 * numElements])
        if sys.byteorder != "big":
            result.byteswap()
        assert len(result) == numElements, result
        return result.tolist()

    def compile(self, ttFont):
        colorRecordIndices, colorRecords = self._compileColorRecords()
        paletteTypes = self._compilePaletteTypes()
        paletteLabels = self._compilePaletteLabels()
        paletteEntryLabels = self._compilePaletteEntryLabels()
        numColorRecords = len(colorRecords) // 4
        offsetToFirstColorRecord = 12 + len(colorRecordIndices)
        if self.version >= 1:
            offsetToFirstColorRecord += 12
        header = struct.pack(
            ">HHHHL",
            self.version,
            self.numPaletteEntries,
            len(self.palettes),
            numColorRecords,
            offsetToFirstColorRecord,
        )
        if self.version == 0:
            dataList = [header, colorRecordIndices, colorRecords]
        else:
            pos = offsetToFirstColorRecord + len(colorRecords)
            if len(paletteTypes) == 0:
                offsetToPaletteTypeArray = 0
            else:
                offsetToPaletteTypeArray = pos
                pos += len(paletteTypes)
            if len(paletteLabels) == 0:
                offsetToPaletteLabelArray = 0
            else:
                offsetToPaletteLabelArray = pos
                pos += len(paletteLabels)
            if len(paletteEntryLabels) == 0:
                offsetToPaletteEntryLabelArray = 0
            else:
                offsetToPaletteEntryLabelArray = pos
                pos += len(paletteLabels)
            header1 = struct.pack(
                ">LLL",
                offsetToPaletteTypeArray,
                offsetToPaletteLabelArray,
                offsetToPaletteEntryLabelArray,
            )
            dataList = [
                header,
                colorRecordIndices,
                header1,
                colorRecords,
                paletteTypes,
                paletteLabels,
                paletteEntryLabels,
            ]
        return bytesjoin(dataList)

    def _compilePalette(self, palette):
        assert len(palette) == self.numPaletteEntries
        pack = lambda c: struct.pack(">BBBB", c.blue, c.green, c.red, c.alpha)
        return bytesjoin([pack(color) for color in palette])

    def _compileColorRecords(self):
        colorRecords, colorRecordIndices, pool = [], [], {}
        for palette in self.palettes:
            packedPalette = self._compilePalette(palette)
            if packedPalette in pool:
                index = pool[packedPalette]
            else:
                index = len(colorRecords)
                colorRecords.append(packedPalette)
                pool[packedPalette] = index
            colorRecordIndices.append(struct.pack(">H", index * self.numPaletteEntries))
        return bytesjoin(colorRecordIndices), bytesjoin(colorRecords)

    def _compilePaletteTypes(self):
        if self.version == 0 or not any(self.paletteTypes):
            return b""
        assert len(self.paletteTypes) == len(self.palettes)
        result = bytesjoin([struct.pack(">I", ptype) for ptype in self.paletteTypes])
        assert len(result) == 4 * len(self.palettes)
        return result

    def _compilePaletteLabels(self):
        if self.version == 0 or all(l == self.NO_NAME_ID for l in self.paletteLabels):
            return b""
        assert len(self.paletteLabels) == len(self.palettes)
        result = bytesjoin([struct.pack(">H", label) for label in self.paletteLabels])
        assert len(result) == 2 * len(self.palettes)
        return result

    def _compilePaletteEntryLabels(self):
        if self.version == 0 or all(
            l == self.NO_NAME_ID for l in self.paletteEntryLabels
        ):
            return b""
        assert len(self.paletteEntryLabels) == self.numPaletteEntries
        result = bytesjoin(
            [struct.pack(">H", label) for label in self.paletteEntryLabels]
        )
        assert len(result) == 2 * self.numPaletteEntries
        return result

    def toXML(self, writer, ttFont):
        numPalettes = len(self.palettes)
        paletteLabels = {i: nameID for (i, nameID) in enumerate(self.paletteLabels)}
        paletteTypes = {i: typ for (i, typ) in enumerate(self.paletteTypes)}
        writer.simpletag("version", value=self.version)
        writer.newline()
        writer.simpletag("numPaletteEntries", value=self.numPaletteEntries)
        writer.newline()
        for index, palette in enumerate(self.palettes):
            attrs = {"index": index}
            paletteType = paletteTypes.get(index, self.DEFAULT_PALETTE_TYPE)
            paletteLabel = paletteLabels.get(index, self.NO_NAME_ID)
            if self.version > 0 and paletteLabel != self.NO_NAME_ID:
                attrs["label"] = paletteLabel
            if self.version > 0 and paletteType != self.DEFAULT_PALETTE_TYPE:
                attrs["type"] = paletteType
            writer.begintag("palette", **attrs)
            writer.newline()
            if (
                self.version > 0
                and paletteLabel != self.NO_NAME_ID
                and ttFont
                and "name" in ttFont
            ):
                name = ttFont["name"].getDebugName(paletteLabel)
                if name is not None:
                    writer.comment(name)
                    writer.newline()
            assert len(palette) == self.numPaletteEntries
            for cindex, color in enumerate(palette):
                color.toXML(writer, ttFont, cindex)
            writer.endtag("palette")
            writer.newline()
        if self.version > 0 and not all(
            l == self.NO_NAME_ID for l in self.paletteEntryLabels
        ):
            writer.begintag("paletteEntryLabels")
            writer.newline()
            for index, label in enumerate(self.paletteEntryLabels):
                if label != self.NO_NAME_ID:
                    writer.simpletag("label", index=index, value=label)
                    if self.version > 0 and label and ttFont and "name" in ttFont:
                        name = ttFont["name"].getDebugName(label)
                        if name is not None:
                            writer.comment(name)
                    writer.newline()
            writer.endtag("paletteEntryLabels")
            writer.newline()

    def fromXML(self, name, attrs, content, ttFont):
        if name == "palette":
            self.paletteLabels.append(int(attrs.get("label", self.NO_NAME_ID)))
            self.paletteTypes.append(int(attrs.get("type", self.DEFAULT_PALETTE_TYPE)))
            palette = []
            for element in content:
                if isinstance(element, str):
                    continue
                attrs = element[1]
                color = Color.fromHex(attrs["value"])
                palette.append(color)
            self.palettes.append(palette)
        elif name == "paletteEntryLabels":
            colorLabels = {}
            for element in content:
                if isinstance(element, str):
                    continue
                elementName, elementAttr, _ = element
                if elementName == "label":
                    labelIndex = safeEval(elementAttr["index"])
                    nameID = safeEval(elementAttr["value"])
                    colorLabels[labelIndex] = nameID
            self.paletteEntryLabels = [
                colorLabels.get(i, self.NO_NAME_ID)
                for i in range(self.numPaletteEntries)
            ]
        elif "value" in attrs:
            value = safeEval(attrs["value"])
            setattr(self, name, value)
            if name == "numPaletteEntries":
                self.paletteEntryLabels = [self.NO_NAME_ID] * self.numPaletteEntries


class Color(namedtuple("Color", "blue green red alpha")):
    def hex(self):
        return "#%02X%02X%02X%02X" % (self.red, self.green, self.blue, self.alpha)

    def __repr__(self):
        return self.hex()

    def toXML(self, writer, ttFont, index=None):
        writer.simpletag("color", value=self.hex(), index=index)
        writer.newline()

    @classmethod
    def fromHex(cls, value):
        if value[0] == "#":
            value = value[1:]
        red = int(value[0:2], 16)
        green = int(value[2:4], 16)
        blue = int(value[4:6], 16)
        alpha = int(value[6:8], 16) if len(value) >= 8 else 0xFF
        return cls(red=red, green=green, blue=blue, alpha=alpha)

    @classmethod
    def fromRGBA(cls, red, green, blue, alpha):
        return cls(red=red, green=green, blue=blue, alpha=alpha)



try:
    from fontTools.misc.symfont import AreaPen
except ImportError:
    AreaPen = None
import unittest
import pytest

precision = 6


def draw1_(pen):
    pen.moveTo((254, 360))
    pen.lineTo((771, 367))
    pen.curveTo((800, 393), (808, 399), (819, 412))
    pen.curveTo((818, 388), (774, 138), (489, 145))
    pen.curveTo((188, 145), (200, 398), (200, 421))
    pen.curveTo((209, 409), (220, 394), (254, 360))
    pen.closePath()


class AreaPenTest(unittest.TestCase):
    @pytest.mark.skipif(AreaPen is None, reason="sympy not installed")
    def test_PScontour_clockwise_line_first(self):
        pen = AreaPen(glyphset=None)
        draw1_(pen)
        self.assertEqual(-104561.35, round(pen.value, precision))

    @pytest.mark.skipif(AreaPen is None, reason="sympy not installed")
    def test_openPaths(self):
        pen = AreaPen()
        pen.moveTo((0, 0))
        pen.endPath()
        self.assertEqual(0, pen.value)

        pen.moveTo((0, 0))
        pen.lineTo((1, 0))
        with self.assertRaises(NotImplementedError):
            pen.endPath()


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from io import BytesIO
from fontTools.ttLib.tables.C_F_F_ import table_C_F_F_


class table_C_F_F__2(table_C_F_F_):
    """Compact Font Format version 2 table

    The ``CFF2`` table contains glyph data for a CFF2-flavored OpenType
    font.

    .. note::
       ``CFF2`` is the successor to ``CFF``, and eliminates much of
       the redundancy incurred by embedding CFF version 1 in an OpenType
       font.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/cff2
    """

    def decompile(self, data, otFont):
        self.cff.decompile(BytesIO(data), otFont, isCFF2=True)
        assert len(self.cff) == 1, "can't deal with multi-font CFF tables."

    def compile(self, otFont):
        f = BytesIO()
        self.cff.compile(f, otFont, isCFF2=True)
        return f.getvalue()



import sys
import os
import datetime
import codecs
import collections
from io import BytesIO
from numbers import Integral
from fontTools.misc import etree
from fontTools.misc import plistlib
from fontTools.misc.textTools import tostr
import pytest
from collections.abc import Mapping


# The testdata is generated using https://github.com/python/cpython/...
# Mac/Tools/plistlib_generate_testdata.py
# which uses PyObjC to control the Cocoa classes for generating plists
datadir = os.path.join(os.path.dirname(__file__), "testdata")
with open(os.path.join(datadir, "test.plist"), "rb") as fp:
    TESTDATA = fp.read()


def _test_pl(use_builtin_types):
    DataClass = bytes if use_builtin_types else plistlib.Data
    pl = dict(
        aString="Doodah",
        aList=["A", "B", 12, 32.5, [1, 2, 3]],
        aFloat=0.5,
        anInt=728,
        aBigInt=2**63 - 44,
        aBigInt2=2**63 + 44,
        aNegativeInt=-5,
        aNegativeBigInt=-80000000000,
        aDict=dict(
            anotherString="<hello & 'hi' there!>",
            aUnicodeValue="M\xe4ssig, Ma\xdf",
            aTrueValue=True,
            aFalseValue=False,
            deeperDict=dict(a=17, b=32.5, c=[1, 2, "text"]),
        ),
        someData=DataClass(b"<binary gunk>"),
        someMoreData=DataClass(b"<lots of binary gunk>\0\1\2\3" * 10),
        nestedData=[DataClass(b"<lots of binary gunk>\0\1\2\3" * 10)],
        aDate=datetime.datetime(2004, 10, 26, 10, 33, 33),
        anEmptyDict=dict(),
        anEmptyList=list(),
    )
    pl["\xc5benraa"] = "That was a unicode key."
    return pl


@pytest.fixture
def pl():
    return _test_pl(use_builtin_types=True)


@pytest.fixture
def pl_no_builtin_types():
    return _test_pl(use_builtin_types=False)


@pytest.fixture(
    params=[True, False],
    ids=["builtin=True", "builtin=False"],
)
def use_builtin_types(request):
    return request.param


@pytest.fixture
def parametrized_pl(use_builtin_types):
    return _test_pl(use_builtin_types), use_builtin_types


def test__test_pl():
    # sanity test that checks that the two values are equivalent
    # (plistlib.Data implements __eq__ against bytes values)
    pl = _test_pl(use_builtin_types=False)
    pl2 = _test_pl(use_builtin_types=True)
    assert pl == pl2


def test_io(tmpdir, parametrized_pl):
    pl, use_builtin_types = parametrized_pl
    testpath = tmpdir / "test.plist"
    with testpath.open("wb") as fp:
        plistlib.dump(pl, fp, use_builtin_types=use_builtin_types)

    with testpath.open("rb") as fp:
        pl2 = plistlib.load(fp, use_builtin_types=use_builtin_types)

    assert pl == pl2

    with pytest.raises(AttributeError):
        plistlib.dump(pl, "filename")

    with pytest.raises(AttributeError):
        plistlib.load("filename")


def test_invalid_type():
    pl = [object()]

    with pytest.raises(TypeError):
        plistlib.dumps(pl)


@pytest.mark.parametrize(
    "pl",
    [
        0,
        2**8 - 1,
        2**8,
        2**16 - 1,
        2**16,
        2**32 - 1,
        2**32,
        2**63 - 1,
        2**64 - 1,
        1,
        -(2**63),
    ],
)
def test_int(pl):
    data = plistlib.dumps(pl)
    pl2 = plistlib.loads(data)
    assert isinstance(pl2, Integral)
    assert pl == pl2
    data2 = plistlib.dumps(pl2)
    assert data == data2


@pytest.mark.parametrize("pl", [2**64 + 1, 2**127 - 1, -(2**64), -(2**127)])
def test_int_overflow(pl):
    with pytest.raises(OverflowError):
        plistlib.dumps(pl)


def test_bytearray(use_builtin_types):
    DataClass = bytes if use_builtin_types else plistlib.Data
    pl = DataClass(b"<binary gunk\0\1\2\3>")
    array = bytearray(pl) if use_builtin_types else bytearray(pl.data)
    data = plistlib.dumps(array)
    pl2 = plistlib.loads(data, use_builtin_types=use_builtin_types)
    assert isinstance(pl2, DataClass)
    assert pl2 == pl
    data2 = plistlib.dumps(pl2, use_builtin_types=use_builtin_types)
    assert data == data2


@pytest.mark.parametrize(
    "DataClass, use_builtin_types",
    [(bytes, True), (plistlib.Data, True), (plistlib.Data, False)],
    ids=[
        "bytes|builtin_types=True",
        "Data|builtin_types=True",
        "Data|builtin_types=False",
    ],
)
def test_bytes_data(DataClass, use_builtin_types):
    pl = DataClass(b"<binary gunk\0\1\2\3>")
    data = plistlib.dumps(pl, use_builtin_types=use_builtin_types)
    pl2 = plistlib.loads(data, use_builtin_types=use_builtin_types)
    assert isinstance(pl2, bytes if use_builtin_types else plistlib.Data)
    assert pl2 == pl
    data2 = plistlib.dumps(pl2, use_builtin_types=use_builtin_types)
    assert data == data2


def test_bytes_string(use_builtin_types):
    pl = b"some ASCII bytes"
    data = plistlib.dumps(pl, use_builtin_types=False)
    pl2 = plistlib.loads(data, use_builtin_types=use_builtin_types)
    assert isinstance(pl2, str)  # it's always a <string>
    assert pl2 == pl.decode()


def test_indentation_array():
    data = [[[[[[[[{"test": "aaaaaa"}]]]]]]]]
    assert plistlib.loads(plistlib.dumps(data)) == data


def test_indentation_dict():
    data = {"1": {"2": {"3": {"4": {"5": {"6": {"7": {"8": {"9": "aaaaaa"}}}}}}}}}
    assert plistlib.loads(plistlib.dumps(data)) == data


def test_indentation_dict_mix():
    data = {"1": {"2": [{"3": [[[[[{"test": "aaaaaa"}]]]]]}]}}
    assert plistlib.loads(plistlib.dumps(data)) == data


@pytest.mark.xfail(reason="we use two spaces, Apple uses tabs")
def test_apple_formatting(parametrized_pl):
    # we also split base64 data into multiple lines differently:
    # both right-justify data to 76 chars, but Apple's treats tabs
    # as 8 spaces, whereas we use 2 spaces
    pl, use_builtin_types = parametrized_pl
    pl = plistlib.loads(TESTDATA, use_builtin_types=use_builtin_types)
    data = plistlib.dumps(pl, use_builtin_types=use_builtin_types)
    assert data == TESTDATA


def test_apple_formatting_fromliteral(parametrized_pl):
    pl, use_builtin_types = parametrized_pl
    pl2 = plistlib.loads(TESTDATA, use_builtin_types=use_builtin_types)
    assert pl == pl2


def test_apple_roundtrips(use_builtin_types):
    pl = plistlib.loads(TESTDATA, use_builtin_types=use_builtin_types)
    data = plistlib.dumps(pl, use_builtin_types=use_builtin_types)
    pl2 = plistlib.loads(data, use_builtin_types=use_builtin_types)
    data2 = plistlib.dumps(pl2, use_builtin_types=use_builtin_types)
    assert data == data2


def test_bytesio(parametrized_pl):
    pl, use_builtin_types = parametrized_pl
    b = BytesIO()
    plistlib.dump(pl, b, use_builtin_types=use_builtin_types)
    pl2 = plistlib.load(BytesIO(b.getvalue()), use_builtin_types=use_builtin_types)
    assert pl == pl2


@pytest.mark.parametrize("sort_keys", [False, True])
def test_keysort_bytesio(sort_keys):
    pl = collections.OrderedDict()
    pl["b"] = 1
    pl["a"] = 2
    pl["c"] = 3

    b = BytesIO()

    plistlib.dump(pl, b, sort_keys=sort_keys)
    pl2 = plistlib.load(BytesIO(b.getvalue()), dict_type=collections.OrderedDict)

    assert dict(pl) == dict(pl2)
    if sort_keys:
        assert list(pl2.keys()) == ["a", "b", "c"]
    else:
        assert list(pl2.keys()) == ["b", "a", "c"]


@pytest.mark.parametrize("sort_keys", [False, True])
def test_keysort(sort_keys):
    pl = collections.OrderedDict()
    pl["b"] = 1
    pl["a"] = 2
    pl["c"] = 3

    data = plistlib.dumps(pl, sort_keys=sort_keys)
    pl2 = plistlib.loads(data, dict_type=collections.OrderedDict)

    assert dict(pl) == dict(pl2)
    if sort_keys:
        assert list(pl2.keys()) == ["a", "b", "c"]
    else:
        assert list(pl2.keys()) == ["b", "a", "c"]


def test_keys_no_string():
    pl = {42: "aNumber"}

    with pytest.raises(TypeError):
        plistlib.dumps(pl)

    b = BytesIO()
    with pytest.raises(TypeError):
        plistlib.dump(pl, b)


def test_skipkeys():
    pl = {42: "aNumber", "snake": "aWord"}

    data = plistlib.dumps(pl, skipkeys=True, sort_keys=False)

    pl2 = plistlib.loads(data)
    assert pl2 == {"snake": "aWord"}

    fp = BytesIO()
    plistlib.dump(pl, fp, skipkeys=True, sort_keys=False)
    data = fp.getvalue()
    pl2 = plistlib.loads(fp.getvalue())
    assert pl2 == {"snake": "aWord"}


def test_tuple_members():
    pl = {"first": (1, 2), "second": (1, 2), "third": (3, 4)}

    data = plistlib.dumps(pl)
    pl2 = plistlib.loads(data)
    assert pl2 == {"first": [1, 2], "second": [1, 2], "third": [3, 4]}
    assert pl2["first"] is not pl2["second"]


def test_list_members():
    pl = {"first": [1, 2], "second": [1, 2], "third": [3, 4]}

    data = plistlib.dumps(pl)
    pl2 = plistlib.loads(data)
    assert pl2 == {"first": [1, 2], "second": [1, 2], "third": [3, 4]}
    assert pl2["first"] is not pl2["second"]


def test_dict_members():
    pl = {"first": {"a": 1}, "second": {"a": 1}, "third": {"b": 2}}

    data = plistlib.dumps(pl)
    pl2 = plistlib.loads(data)
    assert pl2 == {"first": {"a": 1}, "second": {"a": 1}, "third": {"b": 2}}
    assert pl2["first"] is not pl2["second"]


def test_controlcharacters():
    for i in range(128):
        c = chr(i)
        testString = "string containing %s" % c
        if i >= 32 or c in "\r\n\t":
            # \r, \n and \t are the only legal control chars in XML
            data = plistlib.dumps(testString)
            # the stdlib's plistlib writer, as well as the elementtree
            # parser, always replace \r with \n inside string values;
            # lxml doesn't (the ctrl character is escaped), so it roundtrips
            if c != "\r" or etree._have_lxml:
                assert plistlib.loads(data) == testString
        else:
            with pytest.raises(ValueError):
                plistlib.dumps(testString)


def test_non_bmp_characters():
    pl = {"python": "\U0001f40d"}
    data = plistlib.dumps(pl)
    assert plistlib.loads(data) == pl


def test_nondictroot():
    test1 = "abc"
    test2 = [1, 2, 3, "abc"]
    result1 = plistlib.loads(plistlib.dumps(test1))
    result2 = plistlib.loads(plistlib.dumps(test2))
    assert test1 == result1
    assert test2 == result2


def test_invalidarray():
    for i in [
        "<key>key inside an array</key>",
        "<key>key inside an array2</key><real>3</real>",
        "<true/><key>key inside an array3</key>",
    ]:
        with pytest.raises(ValueError):
            plistlib.loads(("<plist><array>%s</array></plist>" % i).encode("utf-8"))


def test_invaliddict():
    for i in [
        "<key><true/>k</key><string>compound key</string>",
        "<key>single key</key>",
        "<string>missing key</string>",
        "<key>k1</key><string>v1</string><real>5.3</real>"
        "<key>k1</key><key>k2</key><string>double key</string>",
    ]:
        with pytest.raises(ValueError):
            plistlib.loads(("<plist><dict>%s</dict></plist>" % i).encode())
        with pytest.raises(ValueError):
            plistlib.loads(
                ("<plist><array><dict>%s</dict></array></plist>" % i).encode()
            )


def test_invalidinteger():
    with pytest.raises(ValueError):
        plistlib.loads(b"<plist><integer>not integer</integer></plist>")


def test_invalidreal():
    with pytest.raises(ValueError):
        plistlib.loads(b"<plist><integer>not real</integer></plist>")


@pytest.mark.parametrize(
    "xml_encoding, encoding, bom",
    [
        (b"utf-8", "utf-8", codecs.BOM_UTF8),
        (b"utf-16", "utf-16-le", codecs.BOM_UTF16_LE),
        (b"utf-16", "utf-16-be", codecs.BOM_UTF16_BE),
        # expat parser (used by ElementTree) does't support UTF-32
        # (b"utf-32", "utf-32-le", codecs.BOM_UTF32_LE),
        # (b"utf-32", "utf-32-be", codecs.BOM_UTF32_BE),
    ],
)
def test_xml_encodings(parametrized_pl, xml_encoding, encoding, bom):
    pl, use_builtin_types = parametrized_pl
    data = TESTDATA.replace(b"UTF-8", xml_encoding)
    data = bom + data.decode("utf-8").encode(encoding)
    pl2 = plistlib.loads(data, use_builtin_types=use_builtin_types)
    assert pl == pl2


def test_fromtree(parametrized_pl):
    pl, use_builtin_types = parametrized_pl
    tree = etree.fromstring(TESTDATA)
    pl2 = plistlib.fromtree(tree, use_builtin_types=use_builtin_types)
    assert pl == pl2


def _strip(txt):
    return (
        "".join(l.strip() for l in tostr(txt, "utf-8").splitlines())
        if txt is not None
        else ""
    )


def test_totree(parametrized_pl):
    pl, use_builtin_types = parametrized_pl
    tree = etree.fromstring(TESTDATA)[0]  # ignore root 'plist' element
    tree2 = plistlib.totree(pl, use_builtin_types=use_builtin_types)
    assert tree.tag == tree2.tag == "dict"
    for (_, e1), (_, e2) in zip(etree.iterwalk(tree), etree.iterwalk(tree2)):
        assert e1.tag == e2.tag
        assert e1.attrib == e2.attrib
        assert len(e1) == len(e2)
        # ignore whitespace
        assert _strip(e1.text) == _strip(e2.text)


def test_no_pretty_print(use_builtin_types):
    data = plistlib.dumps(
        {"data": b"hello" if use_builtin_types else plistlib.Data(b"hello")},
        pretty_print=False,
        use_builtin_types=use_builtin_types,
    )
    assert data == (
        plistlib.XML_DECLARATION + plistlib.PLIST_DOCTYPE + b'<plist version="1.0">'
        b"<dict>"
        b"<key>data</key>"
        b"<data>aGVsbG8=</data>"
        b"</dict>"
        b"</plist>"
    )


def test_readPlist_from_path(pl):
    old_plistlib = pytest.importorskip("fontTools.ufoLib.plistlib")
    path = os.path.join(datadir, "test.plist")
    pl2 = old_plistlib.readPlist(path)
    assert isinstance(pl2["someData"], plistlib.Data)
    assert pl2 == pl


def test_readPlist_from_file(pl):
    old_plistlib = pytest.importorskip("fontTools.ufoLib.plistlib")
    with open(os.path.join(datadir, "test.plist"), "rb") as f:
        pl2 = old_plistlib.readPlist(f)
        assert isinstance(pl2["someData"], plistlib.Data)
        assert pl2 == pl
        assert not f.closed


def test_readPlistFromString(pl):
    old_plistlib = pytest.importorskip("fontTools.ufoLib.plistlib")
    pl2 = old_plistlib.readPlistFromString(TESTDATA)
    assert isinstance(pl2["someData"], plistlib.Data)
    assert pl2 == pl


def test_writePlist_to_path(tmpdir, pl_no_builtin_types):
    old_plistlib = pytest.importorskip("fontTools.ufoLib.plistlib")
    testpath = tmpdir / "test.plist"
    old_plistlib.writePlist(pl_no_builtin_types, str(testpath))
    with testpath.open("rb") as fp:
        pl2 = plistlib.load(fp, use_builtin_types=False)
    assert pl2 == pl_no_builtin_types


def test_writePlist_to_file(tmpdir, pl_no_builtin_types):
    old_plistlib = pytest.importorskip("fontTools.ufoLib.plistlib")
    testpath = tmpdir / "test.plist"
    with testpath.open("wb") as fp:
        old_plistlib.writePlist(pl_no_builtin_types, fp)
    with testpath.open("rb") as fp:
        pl2 = plistlib.load(fp, use_builtin_types=False)
    assert pl2 == pl_no_builtin_types


def test_writePlistToString(pl_no_builtin_types):
    old_plistlib = pytest.importorskip("fontTools.ufoLib.plistlib")
    data = old_plistlib.writePlistToString(pl_no_builtin_types)
    pl2 = plistlib.loads(data)
    assert pl2 == pl_no_builtin_types


def test_load_use_builtin_types_default():
    pl = plistlib.loads(TESTDATA)
    assert isinstance(pl["someData"], bytes)


def test_dump_use_builtin_types_default(pl_no_builtin_types):
    data = plistlib.dumps(pl_no_builtin_types)
    pl2 = plistlib.loads(data)
    assert isinstance(pl2["someData"], bytes)
    assert pl2 == pl_no_builtin_types


def test_non_ascii_bytes():
    with pytest.raises(ValueError, match="invalid non-ASCII bytes"):
        plistlib.dumps("\U0001f40d".encode("utf-8"), use_builtin_types=False)


class CustomMapping(Mapping):
    a = {"a": 1, "b": 2}

    def __getitem__(self, key):
        return self.a[key]

    def __iter__(self):
        return iter(self.a)

    def __len__(self):
        return len(self.a)


def test_custom_mapping():
    test_mapping = CustomMapping()
    data = plistlib.dumps(test_mapping)
    assert plistlib.loads(data) == {"a": 1, "b": 2}


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(sys.argv))



# coding: utf-8
from fontTools.misc.testTools import FakeFont, getXML, parseXML
from fontTools.misc.textTools import deHexStr, hexStr
from fontTools.ttLib import newTable
import unittest


# This is the anchor points table of the first font file in
# “/Library/Fonts/Devanagari Sangam MN.ttc” on macOS 10.12.6.
# For testing, we’ve changed the GlyphIDs to smaller values.
# Also, in the AATLookup, we’ve changed GlyphDataOffset value
# for the end-of-table marker from 0xFFFF to 0 since that is
# what our encoder emits. (The value for end-of-table markers
# does not actually matter).
ANKR_FORMAT_0_DATA = deHexStr(
    "0000 0000 "  #  0: Format=0, Flags=0
    "0000 000C "  #  4: LookupTableOffset=12
    "0000 0024 "  #  8: GlyphDataTableOffset=36
    "0006 0004 0002 "  # 12: LookupFormat=6, UnitSize=4, NUnits=2
    "0008 0001 0000 "  # 18: SearchRange=8, EntrySelector=1, RangeShift=0
    "0001 0000 "  # 24: Glyph=A, Offset=0 (+GlyphDataTableOffset=36)
    "0003 0008 "  # 28: Glyph=C, Offset=8 (+GlyphDataTableOffset=44)
    "FFFF 0000 "  # 32: Glyph=<end>, Offset=<n/a>
    "0000 0001 "  # 36: GlyphData[A].NumPoints=1
    "0235 045E "  # 40: GlyphData[A].Points[0].X=565, .Y=1118
    "0000 0001 "  # 44: GlyphData[C].NumPoints=1
    "FED2 045E "  # 48: GlyphData[C].Points[0].X=-302, .Y=1118
)  # 52: <end>
assert len(ANKR_FORMAT_0_DATA) == 52


ANKR_FORMAT_0_XML = [
    '<AnchorPoints Format="0">',
    '  <Flags value="0"/>',
    "  <Anchors>",
    '    <Lookup glyph="A">',
    "      <!-- AnchorPointCount=1 -->",
    '      <AnchorPoint index="0">',
    '        <XCoordinate value="565"/>',
    '        <YCoordinate value="1118"/>',
    "      </AnchorPoint>",
    "    </Lookup>",
    '    <Lookup glyph="C">',
    "      <!-- AnchorPointCount=1 -->",
    '      <AnchorPoint index="0">',
    '        <XCoordinate value="-302"/>',
    '        <YCoordinate value="1118"/>',
    "      </AnchorPoint>",
    "    </Lookup>",
    "  </Anchors>",
    "</AnchorPoints>",
]


# Same data as ANKR_FORMAT_0_DATA, but with chunks of unused data
# whose presence should not stop us from decompiling the table.
ANKR_FORMAT_0_STRAY_DATA = deHexStr(
    "0000 0000 "  #  0: Format=0, Flags=0
    "0000 0018 "  #  4: LookupTableOffset=24
    "0000 0034 "  #  8: GlyphDataTableOffset=52
    "DEAD BEEF CAFE "  # 12: <stray data>
    "DEAD BEEF CAFE "  # 18: <stray data>
    "0006 0004 0002 "  # 24: LookupFormat=6, UnitSize=4, NUnits=2
    "0008 0001 0000 "  # 30: SearchRange=8, EntrySelector=1, RangeShift=0
    "0001 0000 "  # 36: Glyph=A, Offset=0 (+GlyphDataTableOffset=52)
    "0003 0008 "  # 40: Glyph=C, Offset=8 (+GlyphDataTableOffset=60)
    "FFFF 0000 "  # 44: Glyph=<end>, Offset=<n/a>
    "BEEF F00D "  # 48: <stray data>
    "0000 0001 "  # 52: GlyphData[A].NumPoints=1
    "0235 045E "  # 56: GlyphData[A].Points[0].X=565, .Y=1118
    "0000 0001 "  # 60: GlyphData[C].NumPoints=1
    "FED2 045E "  # 64: GlyphData[C].Points[0].X=-302, .Y=1118
)  # 68: <end>
assert len(ANKR_FORMAT_0_STRAY_DATA) == 68


# Constructed test case where glyphs A and D share the same anchor data.
ANKR_FORMAT_0_SHARING_DATA = deHexStr(
    "0000 0000 "  #  0: Format=0, Flags=0
    "0000 000C "  #  4: LookupTableOffset=12
    "0000 0028 "  #  8: GlyphDataTableOffset=40
    "0006 0004 0003 "  # 12: LookupFormat=6, UnitSize=4, NUnits=3
    "0008 0001 0004 "  # 18: SearchRange=8, EntrySelector=1, RangeShift=4
    "0001 0000 "  # 24: Glyph=A, Offset=0 (+GlyphDataTableOffset=36)
    "0003 0008 "  # 28: Glyph=C, Offset=8 (+GlyphDataTableOffset=44)
    "0004 0000 "  # 32: Glyph=D, Offset=0 (+GlyphDataTableOffset=36)
    "FFFF 0000 "  # 36: Glyph=<end>, Offset=<n/a>
    "0000 0001 "  # 40: GlyphData[A].NumPoints=1
    "0235 045E "  # 44: GlyphData[A].Points[0].X=565, .Y=1118
    "0000 0002 "  # 48: GlyphData[C].NumPoints=2
    "000B 000C "  # 52: GlyphData[C].Points[0].X=11, .Y=12
    "001B 001C "  # 56: GlyphData[C].Points[1].X=27, .Y=28
)  # 60: <end>
assert len(ANKR_FORMAT_0_SHARING_DATA) == 60


ANKR_FORMAT_0_SHARING_XML = [
    '<AnchorPoints Format="0">',
    '  <Flags value="0"/>',
    "  <Anchors>",
    '    <Lookup glyph="A">',
    "      <!-- AnchorPointCount=1 -->",
    '      <AnchorPoint index="0">',
    '        <XCoordinate value="565"/>',
    '        <YCoordinate value="1118"/>',
    "      </AnchorPoint>",
    "    </Lookup>",
    '    <Lookup glyph="C">',
    "      <!-- AnchorPointCount=2 -->",
    '      <AnchorPoint index="0">',
    '        <XCoordinate value="11"/>',
    '        <YCoordinate value="12"/>',
    "      </AnchorPoint>",
    '      <AnchorPoint index="1">',
    '        <XCoordinate value="27"/>',
    '        <YCoordinate value="28"/>',
    "      </AnchorPoint>",
    "    </Lookup>",
    '    <Lookup glyph="D">',
    "      <!-- AnchorPointCount=1 -->",
    '      <AnchorPoint index="0">',
    '        <XCoordinate value="565"/>',
    '        <YCoordinate value="1118"/>',
    "      </AnchorPoint>",
    "    </Lookup>",
    "  </Anchors>",
    "</AnchorPoints>",
]


class ANKRTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        cls.font = FakeFont([".notdef", "A", "B", "C", "D"])

    def decompileToXML(self, data, xml):
        table = newTable("ankr")
        table.decompile(data, self.font)
        self.assertEqual(getXML(table.toXML), xml)

    def compileFromXML(self, xml, data):
        table = newTable("ankr")
        for name, attrs, content in parseXML(xml):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(data))

    def roundtrip(self, data, xml):
        self.decompileToXML(data, xml)
        self.compileFromXML(xml, data)

    def testFormat0(self):
        self.roundtrip(ANKR_FORMAT_0_DATA, ANKR_FORMAT_0_XML)

    def testFormat0_stray(self):
        self.decompileToXML(ANKR_FORMAT_0_STRAY_DATA, ANKR_FORMAT_0_XML)

    def testFormat0_sharing(self):
        self.roundtrip(ANKR_FORMAT_0_SHARING_DATA, ANKR_FORMAT_0_SHARING_XML)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from fontTools import ttLib
from fontTools.ttLib import woff2
from fontTools.ttLib.tables import _g_l_y_f
from fontTools.ttLib.woff2 import (
    WOFF2Reader,
    woff2DirectorySize,
    woff2DirectoryFormat,
    woff2FlagsSize,
    woff2UnknownTagSize,
    woff2Base128MaxSize,
    WOFF2DirectoryEntry,
    getKnownTagIndex,
    packBase128,
    base128Size,
    woff2UnknownTagIndex,
    WOFF2FlavorData,
    woff2TransformedTableTags,
    WOFF2GlyfTable,
    WOFF2LocaTable,
    WOFF2HmtxTable,
    WOFF2Writer,
    unpackBase128,
    unpack255UShort,
    pack255UShort,
)
import unittest
from fontTools.misc import sstruct
from fontTools.misc.textTools import Tag, bytechr, byteord
from fontTools import fontBuilder
from fontTools.pens.ttGlyphPen import TTGlyphPen
from fontTools.pens.recordingPen import RecordingPen
from io import BytesIO
import struct
import os
import random
import copy
from collections import OrderedDict
from functools import partial
import pytest

haveBrotli = False
try:
    try:
        import brotlicffi as brotli
    except ImportError:
        import brotli
    haveBrotli = True
except ImportError:
    pass


# Python 3 renamed 'assertRaisesRegexp' to 'assertRaisesRegex', and fires
# deprecation warnings if a program uses the old name.
if not hasattr(unittest.TestCase, "assertRaisesRegex"):
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp


current_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
data_dir = os.path.join(current_dir, "data")
TTX = os.path.join(data_dir, "TestTTF-Regular.ttx")
OTX = os.path.join(data_dir, "TestOTF-Regular.otx")
METADATA = os.path.join(data_dir, "test_woff2_metadata.xml")

TT_WOFF2 = BytesIO()
CFF_WOFF2 = BytesIO()


def setUpModule():
    if not haveBrotli:
        raise unittest.SkipTest("No module named brotli")
    assert os.path.exists(TTX)
    assert os.path.exists(OTX)
    # import TT-flavoured test font and save it as WOFF2
    ttf = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
    ttf.importXML(TTX)
    ttf.flavor = "woff2"
    ttf.save(TT_WOFF2, reorderTables=None)
    # import CFF-flavoured test font and save it as WOFF2
    otf = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
    otf.importXML(OTX)
    otf.flavor = "woff2"
    otf.save(CFF_WOFF2, reorderTables=None)


class WOFF2ReaderTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.file = BytesIO(CFF_WOFF2.getvalue())
        cls.font = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
        cls.font.importXML(OTX)

    def setUp(self):
        self.file.seek(0)

    def test_bad_signature(self):
        with self.assertRaisesRegex(ttLib.TTLibError, "bad signature"):
            WOFF2Reader(BytesIO(b"wOFF"))

    def test_not_enough_data_header(self):
        incomplete_header = self.file.read(woff2DirectorySize - 1)
        with self.assertRaisesRegex(ttLib.TTLibError, "not enough data"):
            WOFF2Reader(BytesIO(incomplete_header))

    def test_incorrect_compressed_size(self):
        data = self.file.read(woff2DirectorySize)
        header = sstruct.unpack(woff2DirectoryFormat, data)
        header["totalCompressedSize"] = 0
        data = sstruct.pack(woff2DirectoryFormat, header)
        with self.assertRaises((brotli.error, ttLib.TTLibError)):
            WOFF2Reader(BytesIO(data + self.file.read()))

    def test_incorrect_uncompressed_size(self):
        decompress_backup = brotli.decompress
        brotli.decompress = lambda data: b""  # return empty byte string
        with self.assertRaisesRegex(
            ttLib.TTLibError, "unexpected size for decompressed"
        ):
            WOFF2Reader(self.file)
        brotli.decompress = decompress_backup

    def test_incorrect_file_size(self):
        data = self.file.read(woff2DirectorySize)
        header = sstruct.unpack(woff2DirectoryFormat, data)
        header["length"] -= 1
        data = sstruct.pack(woff2DirectoryFormat, header)
        with self.assertRaisesRegex(
            ttLib.TTLibError, "doesn't match the actual file size"
        ):
            WOFF2Reader(BytesIO(data + self.file.read()))

    def test_num_tables(self):
        tags = [t for t in self.font.keys() if t not in ("GlyphOrder", "DSIG")]
        data = self.file.read(woff2DirectorySize)
        header = sstruct.unpack(woff2DirectoryFormat, data)
        self.assertEqual(header["numTables"], len(tags))

    def test_table_tags(self):
        tags = set([t for t in self.font.keys() if t not in ("GlyphOrder", "DSIG")])
        reader = WOFF2Reader(self.file)
        self.assertEqual(set(reader.keys()), tags)

    def test_get_normal_tables(self):
        woff2Reader = WOFF2Reader(self.file)
        specialTags = woff2TransformedTableTags + ("head", "GlyphOrder", "DSIG")
        for tag in [t for t in self.font.keys() if t not in specialTags]:
            origData = self.font.getTableData(tag)
            decompressedData = woff2Reader[tag]
            self.assertEqual(origData, decompressedData)

    def test_reconstruct_unknown(self):
        reader = WOFF2Reader(self.file)
        with self.assertRaisesRegex(ttLib.TTLibError, "transform for table .* unknown"):
            reader.reconstructTable("head")


class WOFF2ReaderTTFTest(WOFF2ReaderTest):
    """Tests specific to TT-flavored fonts."""

    @classmethod
    def setUpClass(cls):
        cls.file = BytesIO(TT_WOFF2.getvalue())
        cls.font = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
        cls.font.importXML(TTX)

    def setUp(self):
        self.file.seek(0)

    def test_reconstruct_glyf(self):
        woff2Reader = WOFF2Reader(self.file)
        reconstructedData = woff2Reader["glyf"]
        self.assertEqual(self.font.getTableData("glyf"), reconstructedData)

    def test_reconstruct_loca(self):
        woff2Reader = WOFF2Reader(self.file)
        reconstructedData = woff2Reader["loca"]
        self.font.getTableData("glyf")  # 'glyf' needs to be compiled before 'loca'
        self.assertEqual(self.font.getTableData("loca"), reconstructedData)
        self.assertTrue(hasattr(woff2Reader.tables["glyf"], "data"))

    def test_reconstruct_loca_not_match_orig_size(self):
        reader = WOFF2Reader(self.file)
        reader.tables["loca"].origLength -= 1
        with self.assertRaisesRegex(
            ttLib.TTLibError, "'loca' table doesn't match original size"
        ):
            reader.reconstructTable("loca")


def normalise_table(font, tag, padding=4):
    """Return normalised table data. Keep 'font' instance unmodified."""
    assert tag in ("glyf", "loca", "head")
    assert tag in font
    if tag == "head":
        origHeadFlags = font["head"].flags
        font["head"].flags |= 1 << 11
        tableData = font["head"].compile(font)
    if font.sfntVersion in ("\x00\x01\x00\x00", "true"):
        assert {"glyf", "loca", "head"}.issubset(font.keys())
        origIndexFormat = font["head"].indexToLocFormat
        if hasattr(font["loca"], "locations"):
            origLocations = font["loca"].locations[:]
        else:
            origLocations = []
        glyfTable = ttLib.newTable("glyf")
        glyfTable.decompile(font.getTableData("glyf"), font)
        glyfTable.padding = padding
        if tag == "glyf":
            tableData = glyfTable.compile(font)
        elif tag == "loca":
            glyfTable.compile(font)
            tableData = font["loca"].compile(font)
        if tag == "head":
            glyfTable.compile(font)
            font["loca"].compile(font)
            tableData = font["head"].compile(font)
        font["head"].indexToLocFormat = origIndexFormat
        font["loca"].set(origLocations)
    if tag == "head":
        font["head"].flags = origHeadFlags
    return tableData


def normalise_font(font, padding=4):
    """Return normalised font data. Keep 'font' instance unmodified."""
    # drop DSIG but keep a copy
    DSIG_copy = copy.deepcopy(font["DSIG"])
    del font["DSIG"]
    # override TTFont attributes
    origFlavor = font.flavor
    origRecalcBBoxes = font.recalcBBoxes
    origRecalcTimestamp = font.recalcTimestamp
    origLazy = font.lazy
    font.flavor = None
    font.recalcBBoxes = False
    font.recalcTimestamp = False
    font.lazy = True
    # save font to temporary stream
    infile = BytesIO()
    font.save(infile)
    infile.seek(0)
    # reorder tables alphabetically
    outfile = BytesIO()
    reader = ttLib.sfnt.SFNTReader(infile)
    writer = ttLib.sfnt.SFNTWriter(
        outfile,
        len(reader.tables),
        reader.sfntVersion,
        reader.flavor,
        reader.flavorData,
    )
    for tag in sorted(reader.keys()):
        if tag in woff2TransformedTableTags + ("head",):
            writer[tag] = normalise_table(font, tag, padding)
        else:
            writer[tag] = reader[tag]
    writer.close()
    # restore font attributes
    font["DSIG"] = DSIG_copy
    font.flavor = origFlavor
    font.recalcBBoxes = origRecalcBBoxes
    font.recalcTimestamp = origRecalcTimestamp
    font.lazy = origLazy
    return outfile.getvalue()


class WOFF2DirectoryEntryTest(unittest.TestCase):
    def setUp(self):
        self.entry = WOFF2DirectoryEntry()

    def test_not_enough_data_table_flags(self):
        with self.assertRaisesRegex(ttLib.TTLibError, "can't read table 'flags'"):
            self.entry.fromString(b"")

    def test_not_enough_data_table_tag(self):
        incompleteData = bytearray([0x3F, 0, 0, 0])
        with self.assertRaisesRegex(ttLib.TTLibError, "can't read table 'tag'"):
            self.entry.fromString(bytes(incompleteData))

    def test_loca_zero_transformLength(self):
        data = bytechr(getKnownTagIndex("loca"))  # flags
        data += packBase128(random.randint(1, 100))  # origLength
        data += packBase128(1)  # non-zero transformLength
        with self.assertRaisesRegex(
            ttLib.TTLibError, "transformLength of the 'loca' table must be 0"
        ):
            self.entry.fromString(data)

    def test_fromFile(self):
        unknownTag = Tag("ZZZZ")
        data = bytechr(getKnownTagIndex(unknownTag))
        data += unknownTag.tobytes()
        data += packBase128(random.randint(1, 100))
        expectedPos = len(data)
        f = BytesIO(data + b"\0" * 100)
        self.entry.fromFile(f)
        self.assertEqual(f.tell(), expectedPos)

    def test_transformed_toString(self):
        self.entry.tag = Tag("glyf")
        self.entry.flags = getKnownTagIndex(self.entry.tag)
        self.entry.origLength = random.randint(101, 200)
        self.entry.length = random.randint(1, 100)
        expectedSize = (
            woff2FlagsSize
            + base128Size(self.entry.origLength)
            + base128Size(self.entry.length)
        )
        data = self.entry.toString()
        self.assertEqual(len(data), expectedSize)

    def test_known_toString(self):
        self.entry.tag = Tag("head")
        self.entry.flags = getKnownTagIndex(self.entry.tag)
        self.entry.origLength = 54
        expectedSize = woff2FlagsSize + base128Size(self.entry.origLength)
        data = self.entry.toString()
        self.assertEqual(len(data), expectedSize)

    def test_unknown_toString(self):
        self.entry.tag = Tag("ZZZZ")
        self.entry.flags = woff2UnknownTagIndex
        self.entry.origLength = random.randint(1, 100)
        expectedSize = (
            woff2FlagsSize + woff2UnknownTagSize + base128Size(self.entry.origLength)
        )
        data = self.entry.toString()
        self.assertEqual(len(data), expectedSize)

    def test_glyf_loca_transform_flags(self):
        for tag in ("glyf", "loca"):
            entry = WOFF2DirectoryEntry()
            entry.tag = Tag(tag)
            entry.flags = getKnownTagIndex(entry.tag)

            self.assertEqual(entry.transformVersion, 0)
            self.assertTrue(entry.transformed)

            entry.transformed = False

            self.assertEqual(entry.transformVersion, 3)
            self.assertEqual(entry.flags & 0b11000000, (3 << 6))
            self.assertFalse(entry.transformed)

    def test_other_transform_flags(self):
        entry = WOFF2DirectoryEntry()
        entry.tag = Tag("ZZZZ")
        entry.flags = woff2UnknownTagIndex

        self.assertEqual(entry.transformVersion, 0)
        self.assertFalse(entry.transformed)

        entry.transformed = True

        self.assertEqual(entry.transformVersion, 1)
        self.assertEqual(entry.flags & 0b11000000, (1 << 6))
        self.assertTrue(entry.transformed)


class DummyReader(WOFF2Reader):
    def __init__(self, file, checkChecksums=1, fontNumber=-1):
        self.file = file
        for attr in (
            "majorVersion",
            "minorVersion",
            "metaOffset",
            "metaLength",
            "metaOrigLength",
            "privLength",
            "privOffset",
        ):
            setattr(self, attr, 0)
        self.tables = {}


class WOFF2FlavorDataTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        assert os.path.exists(METADATA)
        with open(METADATA, "rb") as f:
            cls.xml_metadata = f.read()
        cls.compressed_metadata = brotli.compress(
            cls.xml_metadata, mode=brotli.MODE_TEXT
        )
        # make random byte strings; font data must be 4-byte aligned
        cls.fontdata = bytes(bytearray(random.sample(range(0, 256), 80)))
        cls.privData = bytes(bytearray(random.sample(range(0, 256), 20)))

    def setUp(self):
        self.file = BytesIO(self.fontdata)
        self.file.seek(0, 2)

    def test_get_metaData_no_privData(self):
        self.file.write(self.compressed_metadata)
        reader = DummyReader(self.file)
        reader.metaOffset = len(self.fontdata)
        reader.metaLength = len(self.compressed_metadata)
        reader.metaOrigLength = len(self.xml_metadata)
        flavorData = WOFF2FlavorData(reader)
        self.assertEqual(self.xml_metadata, flavorData.metaData)

    def test_get_privData_no_metaData(self):
        self.file.write(self.privData)
        reader = DummyReader(self.file)
        reader.privOffset = len(self.fontdata)
        reader.privLength = len(self.privData)
        flavorData = WOFF2FlavorData(reader)
        self.assertEqual(self.privData, flavorData.privData)

    def test_get_metaData_and_privData(self):
        self.file.write(self.compressed_metadata + self.privData)
        reader = DummyReader(self.file)
        reader.metaOffset = len(self.fontdata)
        reader.metaLength = len(self.compressed_metadata)
        reader.metaOrigLength = len(self.xml_metadata)
        reader.privOffset = reader.metaOffset + reader.metaLength
        reader.privLength = len(self.privData)
        flavorData = WOFF2FlavorData(reader)
        self.assertEqual(self.xml_metadata, flavorData.metaData)
        self.assertEqual(self.privData, flavorData.privData)

    def test_get_major_minorVersion(self):
        reader = DummyReader(self.file)
        reader.majorVersion = reader.minorVersion = 1
        flavorData = WOFF2FlavorData(reader)
        self.assertEqual(flavorData.majorVersion, 1)
        self.assertEqual(flavorData.minorVersion, 1)

    def test_mutually_exclusive_args(self):
        msg = "arguments are mutually exclusive"
        reader = DummyReader(self.file)
        with self.assertRaisesRegex(TypeError, msg):
            WOFF2FlavorData(reader, transformedTables={"hmtx"})
        with self.assertRaisesRegex(TypeError, msg):
            WOFF2FlavorData(reader, data=WOFF2FlavorData())

    def test_transformedTables_default(self):
        flavorData = WOFF2FlavorData()
        self.assertEqual(flavorData.transformedTables, set(woff2TransformedTableTags))

    def test_transformedTables_invalid(self):
        msg = r"'glyf' and 'loca' must be transformed \(or not\) together"

        with self.assertRaisesRegex(ValueError, msg):
            WOFF2FlavorData(transformedTables={"glyf"})

        with self.assertRaisesRegex(ValueError, msg):
            WOFF2FlavorData(transformedTables={"loca"})


class WOFF2WriterTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.font = ttLib.TTFont(
            recalcBBoxes=False, recalcTimestamp=False, flavor="woff2"
        )
        cls.font.importXML(OTX)
        cls.tags = sorted(t for t in cls.font.keys() if t != "GlyphOrder")
        cls.numTables = len(cls.tags)
        cls.file = BytesIO(CFF_WOFF2.getvalue())
        cls.file.seek(0, 2)
        cls.length = (cls.file.tell() + 3) & ~3
        cls.setUpFlavorData()

    @classmethod
    def setUpFlavorData(cls):
        assert os.path.exists(METADATA)
        with open(METADATA, "rb") as f:
            cls.xml_metadata = f.read()
        cls.compressed_metadata = brotli.compress(
            cls.xml_metadata, mode=brotli.MODE_TEXT
        )
        cls.privData = bytes(bytearray(random.sample(range(0, 256), 20)))

    def setUp(self):
        self.file.seek(0)
        self.writer = WOFF2Writer(BytesIO(), self.numTables, self.font.sfntVersion)

    def test_DSIG_dropped(self):
        self.writer["DSIG"] = b"\0"
        self.assertEqual(len(self.writer.tables), 0)
        self.assertEqual(self.writer.numTables, self.numTables - 1)

    def test_no_rewrite_table(self):
        self.writer["ZZZZ"] = b"\0"
        with self.assertRaisesRegex(ttLib.TTLibError, "cannot rewrite"):
            self.writer["ZZZZ"] = b"\0"

    def test_num_tables(self):
        self.writer["ABCD"] = b"\0"
        with self.assertRaisesRegex(ttLib.TTLibError, "wrong number of tables"):
            self.writer.close()

    def test_required_tables(self):
        font = ttLib.TTFont(flavor="woff2")
        with self.assertRaisesRegex(ttLib.TTLibError, "missing required table"):
            font.save(BytesIO())

    def test_head_transform_flag(self):
        headData = self.font.getTableData("head")
        origFlags = byteord(headData[16])
        woff2font = ttLib.TTFont(self.file)
        newHeadData = woff2font.getTableData("head")
        modifiedFlags = byteord(newHeadData[16])
        self.assertNotEqual(origFlags, modifiedFlags)
        restoredFlags = modifiedFlags & ~0x08  # turn off bit 11
        self.assertEqual(origFlags, restoredFlags)

    def test_tables_sorted_alphabetically(self):
        expected = sorted([t for t in self.tags if t != "DSIG"])
        woff2font = ttLib.TTFont(self.file)
        self.assertEqual(expected, list(woff2font.reader.keys()))

    def test_checksums(self):
        normFile = BytesIO(normalise_font(self.font, padding=4))
        normFile.seek(0)
        normFont = ttLib.TTFont(normFile, checkChecksums=2)
        w2font = ttLib.TTFont(self.file)
        # force reconstructing glyf table using 4-byte padding
        w2font.reader.padding = 4
        for tag in [t for t in self.tags if t != "DSIG"]:
            w2data = w2font.reader[tag]
            normData = normFont.reader[tag]
            if tag == "head":
                w2data = w2data[:8] + b"\0\0\0\0" + w2data[12:]
                normData = normData[:8] + b"\0\0\0\0" + normData[12:]
            w2CheckSum = ttLib.sfnt.calcChecksum(w2data)
            normCheckSum = ttLib.sfnt.calcChecksum(normData)
            self.assertEqual(w2CheckSum, normCheckSum)
        normCheckSumAdjustment = normFont["head"].checkSumAdjustment
        self.assertEqual(normCheckSumAdjustment, w2font["head"].checkSumAdjustment)

    def test_calcSFNTChecksumsLengthsAndOffsets(self):
        normFont = ttLib.TTFont(BytesIO(normalise_font(self.font, padding=4)))
        for tag in self.tags:
            self.writer[tag] = self.font.getTableData(tag)
        self.writer._normaliseGlyfAndLoca(padding=4)
        self.writer._setHeadTransformFlag()
        self.writer.tables = OrderedDict(sorted(self.writer.tables.items()))
        self.writer._calcSFNTChecksumsLengthsAndOffsets()
        for tag, entry in normFont.reader.tables.items():
            self.assertEqual(entry.offset, self.writer.tables[tag].origOffset)
            self.assertEqual(entry.length, self.writer.tables[tag].origLength)
            self.assertEqual(entry.checkSum, self.writer.tables[tag].checkSum)

    def test_bad_sfntVersion(self):
        for i in range(self.numTables):
            self.writer[bytechr(65 + i) * 4] = b"\0"
        self.writer.sfntVersion = "ZZZZ"
        with self.assertRaisesRegex(ttLib.TTLibError, "bad sfntVersion"):
            self.writer.close()

    def test_calcTotalSize_no_flavorData(self):
        expected = self.length
        self.writer.file = BytesIO()
        for tag in self.tags:
            self.writer[tag] = self.font.getTableData(tag)
        self.writer.close()
        self.assertEqual(expected, self.writer.length)
        self.assertEqual(expected, self.writer.file.tell())

    def test_calcTotalSize_with_metaData(self):
        expected = self.length + len(self.compressed_metadata)
        flavorData = self.writer.flavorData = WOFF2FlavorData()
        flavorData.metaData = self.xml_metadata
        self.writer.file = BytesIO()
        for tag in self.tags:
            self.writer[tag] = self.font.getTableData(tag)
        self.writer.close()
        self.assertEqual(expected, self.writer.length)
        self.assertEqual(expected, self.writer.file.tell())

    def test_calcTotalSize_with_privData(self):
        expected = self.length + len(self.privData)
        flavorData = self.writer.flavorData = WOFF2FlavorData()
        flavorData.privData = self.privData
        self.writer.file = BytesIO()
        for tag in self.tags:
            self.writer[tag] = self.font.getTableData(tag)
        self.writer.close()
        self.assertEqual(expected, self.writer.length)
        self.assertEqual(expected, self.writer.file.tell())

    def test_calcTotalSize_with_metaData_and_privData(self):
        metaDataLength = (len(self.compressed_metadata) + 3) & ~3
        expected = self.length + metaDataLength + len(self.privData)
        flavorData = self.writer.flavorData = WOFF2FlavorData()
        flavorData.metaData = self.xml_metadata
        flavorData.privData = self.privData
        self.writer.file = BytesIO()
        for tag in self.tags:
            self.writer[tag] = self.font.getTableData(tag)
        self.writer.close()
        self.assertEqual(expected, self.writer.length)
        self.assertEqual(expected, self.writer.file.tell())

    def test_getVersion(self):
        # no version
        self.assertEqual((0, 0), self.writer._getVersion())
        # version from head.fontRevision
        fontRevision = self.font["head"].fontRevision
        versionTuple = tuple(int(i) for i in str(fontRevision).split("."))
        entry = self.writer.tables["head"] = ttLib.newTable("head")
        entry.data = self.font.getTableData("head")
        self.assertEqual(versionTuple, self.writer._getVersion())
        # version from writer.flavorData
        flavorData = self.writer.flavorData = WOFF2FlavorData()
        flavorData.majorVersion, flavorData.minorVersion = (10, 11)
        self.assertEqual((10, 11), self.writer._getVersion())

    def test_hmtx_trasform(self):
        tableTransforms = {"glyf", "loca", "hmtx"}

        writer = WOFF2Writer(BytesIO(), self.numTables, self.font.sfntVersion)
        writer.flavorData = WOFF2FlavorData(transformedTables=tableTransforms)

        for tag in self.tags:
            writer[tag] = self.font.getTableData(tag)
        writer.close()

        # enabling hmtx transform has no effect when font has no glyf table
        self.assertEqual(writer.file.getvalue(), CFF_WOFF2.getvalue())

    def test_no_transforms(self):
        writer = WOFF2Writer(BytesIO(), self.numTables, self.font.sfntVersion)
        writer.flavorData = WOFF2FlavorData(transformedTables=())

        for tag in self.tags:
            writer[tag] = self.font.getTableData(tag)
        writer.close()

        # transforms settings have no effect when font is CFF-flavored, since
        # all the current transforms only apply to TrueType-flavored fonts.
        self.assertEqual(writer.file.getvalue(), CFF_WOFF2.getvalue())


class WOFF2WriterTTFTest(WOFF2WriterTest):
    @classmethod
    def setUpClass(cls):
        cls.font = ttLib.TTFont(
            recalcBBoxes=False, recalcTimestamp=False, flavor="woff2"
        )
        cls.font.importXML(TTX)
        cls.tags = sorted(t for t in cls.font.keys() if t != "GlyphOrder")
        cls.numTables = len(cls.tags)
        cls.file = BytesIO(TT_WOFF2.getvalue())
        cls.file.seek(0, 2)
        cls.length = (cls.file.tell() + 3) & ~3
        cls.setUpFlavorData()

    def test_normaliseGlyfAndLoca(self):
        normTables = {}
        for tag in ("head", "loca", "glyf"):
            normTables[tag] = normalise_table(self.font, tag, padding=4)
        for tag in self.tags:
            tableData = self.font.getTableData(tag)
            self.writer[tag] = tableData
            if tag in normTables:
                self.assertNotEqual(tableData, normTables[tag])
        self.writer._normaliseGlyfAndLoca(padding=4)
        self.writer._setHeadTransformFlag()
        for tag in normTables:
            self.assertEqual(self.writer.tables[tag].data, normTables[tag])

    def test_hmtx_trasform(self):
        def compile_hmtx(compressed):
            tableTransforms = woff2TransformedTableTags
            if compressed:
                tableTransforms += ("hmtx",)
            writer = WOFF2Writer(BytesIO(), self.numTables, self.font.sfntVersion)
            writer.flavorData = WOFF2FlavorData(transformedTables=tableTransforms)
            for tag in self.tags:
                writer[tag] = self.font.getTableData(tag)
            writer.close()
            return writer.tables["hmtx"].length

        uncompressed_length = compile_hmtx(compressed=False)
        compressed_length = compile_hmtx(compressed=True)

        # enabling optional hmtx transform shaves off a few bytes
        self.assertLess(compressed_length, uncompressed_length)

    def test_no_transforms(self):
        writer = WOFF2Writer(BytesIO(), self.numTables, self.font.sfntVersion)
        writer.flavorData = WOFF2FlavorData(transformedTables=())

        for tag in self.tags:
            writer[tag] = self.font.getTableData(tag)
        writer.close()

        self.assertNotEqual(writer.file.getvalue(), TT_WOFF2.getvalue())

        writer.file.seek(0)
        reader = WOFF2Reader(writer.file)
        self.assertEqual(len(reader.flavorData.transformedTables), 0)


class WOFF2LocaTableTest(unittest.TestCase):
    def setUp(self):
        self.font = font = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
        font["head"] = ttLib.newTable("head")
        font["loca"] = WOFF2LocaTable()
        font["glyf"] = WOFF2GlyfTable()

    def test_compile_short_loca(self):
        locaTable = self.font["loca"]
        locaTable.set(list(range(0, 0x20000, 2)))
        self.font["glyf"].indexFormat = 0
        locaData = locaTable.compile(self.font)
        self.assertEqual(len(locaData), 0x20000)

    def test_compile_short_loca_overflow(self):
        locaTable = self.font["loca"]
        locaTable.set(list(range(0x20000 + 1)))
        self.font["glyf"].indexFormat = 0
        with self.assertRaisesRegex(
            ttLib.TTLibError, "indexFormat is 0 but local offsets > 0x20000"
        ):
            locaTable.compile(self.font)

    def test_compile_short_loca_not_multiples_of_2(self):
        locaTable = self.font["loca"]
        locaTable.set([1, 3, 5, 7])
        self.font["glyf"].indexFormat = 0
        with self.assertRaisesRegex(ttLib.TTLibError, "offsets not multiples of 2"):
            locaTable.compile(self.font)

    def test_compile_long_loca(self):
        locaTable = self.font["loca"]
        locaTable.set(list(range(0x20001)))
        self.font["glyf"].indexFormat = 1
        locaData = locaTable.compile(self.font)
        self.assertEqual(len(locaData), 0x20001 * 4)

    def test_compile_set_indexToLocFormat_0(self):
        locaTable = self.font["loca"]
        # offsets are all multiples of 2 and max length is < 0x10000
        locaTable.set(list(range(0, 0x20000, 2)))
        locaTable.compile(self.font)
        newIndexFormat = self.font["head"].indexToLocFormat
        self.assertEqual(0, newIndexFormat)

    def test_compile_set_indexToLocFormat_1(self):
        locaTable = self.font["loca"]
        # offsets are not multiples of 2
        locaTable.set(list(range(10)))
        locaTable.compile(self.font)
        newIndexFormat = self.font["head"].indexToLocFormat
        self.assertEqual(1, newIndexFormat)
        # max length is >= 0x10000
        locaTable.set(list(range(0, 0x20000 + 1, 2)))
        locaTable.compile(self.font)
        newIndexFormat = self.font["head"].indexToLocFormat
        self.assertEqual(1, newIndexFormat)


class WOFF2GlyfTableTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        font = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
        font.importXML(TTX)
        cls.tables = {}
        cls.transformedTags = ("maxp", "head", "loca", "glyf")
        for tag in reversed(cls.transformedTags):  # compile in inverse order
            cls.tables[tag] = font.getTableData(tag)
        infile = BytesIO(TT_WOFF2.getvalue())
        reader = WOFF2Reader(infile)
        cls.transformedGlyfData = reader.tables["glyf"].loadData(reader.transformBuffer)
        cls.glyphOrder = [".notdef"] + [
            "glyph%.5d" % i for i in range(1, font["maxp"].numGlyphs)
        ]

    def setUp(self):
        self.font = font = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
        font.setGlyphOrder(self.glyphOrder)
        font["head"] = ttLib.newTable("head")
        font["maxp"] = ttLib.newTable("maxp")
        font["loca"] = WOFF2LocaTable()
        font["glyf"] = WOFF2GlyfTable()
        for tag in self.transformedTags:
            font[tag].decompile(self.tables[tag], font)

    def test_reconstruct_glyf_padded_4(self):
        glyfTable = WOFF2GlyfTable()
        glyfTable.reconstruct(self.transformedGlyfData, self.font)
        glyfTable.padding = 4
        data = glyfTable.compile(self.font)
        normGlyfData = normalise_table(self.font, "glyf", glyfTable.padding)
        self.assertEqual(normGlyfData, data)

    def test_reconstruct_glyf_padded_2(self):
        glyfTable = WOFF2GlyfTable()
        glyfTable.reconstruct(self.transformedGlyfData, self.font)
        glyfTable.padding = 2
        data = glyfTable.compile(self.font)
        normGlyfData = normalise_table(self.font, "glyf", glyfTable.padding)
        self.assertEqual(normGlyfData, data)

    def test_reconstruct_glyf_unpadded(self):
        glyfTable = WOFF2GlyfTable()
        glyfTable.reconstruct(self.transformedGlyfData, self.font)
        data = glyfTable.compile(self.font)
        self.assertEqual(self.tables["glyf"], data)

    def test_reconstruct_glyf_incorrect_glyphOrder(self):
        glyfTable = WOFF2GlyfTable()
        badGlyphOrder = self.font.getGlyphOrder()[:-1]
        self.font.setGlyphOrder(badGlyphOrder)
        with self.assertRaisesRegex(ttLib.TTLibError, "incorrect glyphOrder"):
            glyfTable.reconstruct(self.transformedGlyfData, self.font)

    def test_reconstruct_glyf_missing_glyphOrder(self):
        glyfTable = WOFF2GlyfTable()
        del self.font.glyphOrder
        numGlyphs = self.font["maxp"].numGlyphs
        del self.font["maxp"]
        glyfTable.reconstruct(self.transformedGlyfData, self.font)
        expected = [".notdef"]
        expected.extend(["glyph%.5d" % i for i in range(1, numGlyphs)])
        self.assertEqual(expected, glyfTable.glyphOrder)

    def test_reconstruct_loca_padded_4(self):
        locaTable = self.font["loca"] = WOFF2LocaTable()
        glyfTable = self.font["glyf"] = WOFF2GlyfTable()
        glyfTable.reconstruct(self.transformedGlyfData, self.font)
        glyfTable.padding = 4
        glyfTable.compile(self.font)
        data = locaTable.compile(self.font)
        normLocaData = normalise_table(self.font, "loca", glyfTable.padding)
        self.assertEqual(normLocaData, data)

    def test_reconstruct_loca_padded_2(self):
        locaTable = self.font["loca"] = WOFF2LocaTable()
        glyfTable = self.font["glyf"] = WOFF2GlyfTable()
        glyfTable.reconstruct(self.transformedGlyfData, self.font)
        glyfTable.padding = 2
        glyfTable.compile(self.font)
        data = locaTable.compile(self.font)
        normLocaData = normalise_table(self.font, "loca", glyfTable.padding)
        self.assertEqual(normLocaData, data)

    def test_reconstruct_loca_unpadded(self):
        locaTable = self.font["loca"] = WOFF2LocaTable()
        glyfTable = self.font["glyf"] = WOFF2GlyfTable()
        glyfTable.reconstruct(self.transformedGlyfData, self.font)
        glyfTable.compile(self.font)
        data = locaTable.compile(self.font)
        self.assertEqual(self.tables["loca"], data)

    def test_reconstruct_glyf_header_not_enough_data(self):
        with self.assertRaisesRegex(ttLib.TTLibError, "not enough 'glyf' data"):
            WOFF2GlyfTable().reconstruct(b"", self.font)

    def test_reconstruct_glyf_table_incorrect_size(self):
        msg = "incorrect size of transformed 'glyf'"
        with self.assertRaisesRegex(ttLib.TTLibError, msg):
            WOFF2GlyfTable().reconstruct(self.transformedGlyfData + b"\x00", self.font)
        with self.assertRaisesRegex(ttLib.TTLibError, msg):
            WOFF2GlyfTable().reconstruct(self.transformedGlyfData[:-1], self.font)

    def test_transform_glyf(self):
        glyfTable = self.font["glyf"]
        data = glyfTable.transform(self.font)
        self.assertEqual(self.transformedGlyfData, data)

    def test_roundtrip_glyf_reconstruct_and_transform(self):
        glyfTable = WOFF2GlyfTable()
        glyfTable.reconstruct(self.transformedGlyfData, self.font)
        data = glyfTable.transform(self.font)
        self.assertEqual(self.transformedGlyfData, data)

    def test_roundtrip_glyf_transform_and_reconstruct(self):
        glyfTable = self.font["glyf"]
        transformedData = glyfTable.transform(self.font)
        newGlyfTable = WOFF2GlyfTable()
        newGlyfTable.reconstruct(transformedData, self.font)
        newGlyfTable.padding = 4
        reconstructedData = newGlyfTable.compile(self.font)
        normGlyfData = normalise_table(self.font, "glyf", newGlyfTable.padding)
        self.assertEqual(normGlyfData, reconstructedData)


@pytest.fixture(scope="module")
def fontfile():
    class Glyph(object):
        def __init__(self, empty=False, **kwargs):
            if not empty:
                self.draw = partial(self.drawRect, **kwargs)
            else:
                self.draw = lambda pen: None

        @staticmethod
        def drawRect(pen, xMin, xMax):
            pen.moveTo((xMin, 0))
            pen.lineTo((xMin, 1000))
            pen.lineTo((xMax, 1000))
            pen.lineTo((xMax, 0))
            pen.closePath()

    class CompositeGlyph(object):
        def __init__(self, components):
            self.components = components

        def draw(self, pen):
            for baseGlyph, (offsetX, offsetY) in self.components:
                pen.addComponent(baseGlyph, (1, 0, 0, 1, offsetX, offsetY))

    fb = fontBuilder.FontBuilder(unitsPerEm=1000, isTTF=True)
    fb.setupGlyphOrder(
        [".notdef", "space", "A", "acutecomb", "Aacute", "zero", "one", "two"]
    )
    fb.setupCharacterMap(
        {
            0x20: "space",
            0x41: "A",
            0x0301: "acutecomb",
            0xC1: "Aacute",
            0x30: "zero",
            0x31: "one",
            0x32: "two",
        }
    )
    fb.setupHorizontalMetrics(
        {
            ".notdef": (500, 50),
            "space": (600, 0),
            "A": (550, 40),
            "acutecomb": (0, -40),
            "Aacute": (550, 40),
            "zero": (500, 30),
            "one": (500, 50),
            "two": (500, 40),
        }
    )
    fb.setupHorizontalHeader(ascent=1000, descent=-200)

    srcGlyphs = {
        ".notdef": Glyph(xMin=50, xMax=450),
        "space": Glyph(empty=True),
        "A": Glyph(xMin=40, xMax=510),
        "acutecomb": Glyph(xMin=-40, xMax=60),
        "Aacute": CompositeGlyph([("A", (0, 0)), ("acutecomb", (200, 0))]),
        "zero": Glyph(xMin=30, xMax=470),
        "one": Glyph(xMin=50, xMax=450),
        "two": Glyph(xMin=40, xMax=460),
    }
    pen = TTGlyphPen(srcGlyphs)
    glyphSet = {}
    for glyphName, glyph in srcGlyphs.items():
        glyph.draw(pen)
        glyphSet[glyphName] = pen.glyph()
    fb.setupGlyf(glyphSet)

    fb.setupNameTable(
        {
            "familyName": "TestWOFF2",
            "styleName": "Regular",
            "uniqueFontIdentifier": "TestWOFF2 Regular; Version 1.000; ABCD",
            "fullName": "TestWOFF2 Regular",
            "version": "Version 1.000",
            "psName": "TestWOFF2-Regular",
        }
    )
    fb.setupOS2()
    fb.setupPost()

    buf = BytesIO()
    fb.save(buf)
    buf.seek(0)

    assert fb.font["maxp"].numGlyphs == 8
    assert fb.font["hhea"].numberOfHMetrics == 6
    for glyphName in fb.font.getGlyphOrder():
        xMin = getattr(fb.font["glyf"][glyphName], "xMin", 0)
        assert xMin == fb.font["hmtx"][glyphName][1]

    return buf


@pytest.fixture
def ttFont(fontfile):
    return ttLib.TTFont(fontfile, recalcBBoxes=False, recalcTimestamp=False)


class WOFF2HmtxTableTest(object):
    def test_transform_no_sidebearings(self, ttFont):
        hmtxTable = WOFF2HmtxTable()
        hmtxTable.metrics = ttFont["hmtx"].metrics

        data = hmtxTable.transform(ttFont)

        assert data == (
            b"\x03"  # 00000011 | bits 0 and 1 are set (no sidebearings arrays)
            # advanceWidthArray
            b"\x01\xf4"  # .notdef: 500
            b"\x02X"  # space: 600
            b"\x02&"  # A: 550
            b"\x00\x00"  # acutecomb: 0
            b"\x02&"  # Aacute: 550
            b"\x01\xf4"  # zero: 500
        )

    def test_transform_proportional_sidebearings(self, ttFont):
        hmtxTable = WOFF2HmtxTable()
        metrics = ttFont["hmtx"].metrics
        # force one of the proportional glyphs to have its left sidebearing be
        # different from its xMin (40)
        metrics["A"] = (550, 39)
        hmtxTable.metrics = metrics

        assert ttFont["glyf"]["A"].xMin != metrics["A"][1]

        data = hmtxTable.transform(ttFont)

        assert data == (
            b"\x02"  # 00000010 | bits 0 unset: explicit proportional sidebearings
            # advanceWidthArray
            b"\x01\xf4"  # .notdef: 500
            b"\x02X"  # space: 600
            b"\x02&"  # A: 550
            b"\x00\x00"  # acutecomb: 0
            b"\x02&"  # Aacute: 550
            b"\x01\xf4"  # zero: 500
            # lsbArray
            b"\x002"  # .notdef: 50
            b"\x00\x00"  # space: 0
            b"\x00'"  # A: 39 (xMin: 40)
            b"\xff\xd8"  # acutecomb: -40
            b"\x00("  # Aacute: 40
            b"\x00\x1e"  # zero: 30
        )

    def test_transform_monospaced_sidebearings(self, ttFont):
        hmtxTable = WOFF2HmtxTable()
        metrics = ttFont["hmtx"].metrics
        hmtxTable.metrics = metrics

        # force one of the monospaced glyphs at the end of hmtx table to have
        # its xMin different from its left sidebearing (50)
        ttFont["glyf"]["one"].xMin = metrics["one"][1] + 1

        data = hmtxTable.transform(ttFont)

        assert data == (
            b"\x01"  # 00000001 | bits 1 unset: explicit monospaced sidebearings
            # advanceWidthArray
            b"\x01\xf4"  # .notdef: 500
            b"\x02X"  # space: 600
            b"\x02&"  # A: 550
            b"\x00\x00"  # acutecomb: 0
            b"\x02&"  # Aacute: 550
            b"\x01\xf4"  # zero: 500
            # leftSideBearingArray
            b"\x002"  # one: 50 (xMin: 51)
            b"\x00("  # two: 40
        )

    def test_transform_not_applicable(self, ttFont):
        hmtxTable = WOFF2HmtxTable()
        metrics = ttFont["hmtx"].metrics
        # force both a proportional and monospaced glyph to have sidebearings
        # different from the respective xMin coordinates
        metrics["A"] = (550, 39)
        metrics["one"] = (500, 51)
        hmtxTable.metrics = metrics

        # 'None' signals to fall back using untransformed hmtx table data
        assert hmtxTable.transform(ttFont) is None

    def test_reconstruct_no_sidebearings(self, ttFont):
        hmtxTable = WOFF2HmtxTable()

        data = (
            b"\x03"  # 00000011 | bits 0 and 1 are set (no sidebearings arrays)
            # advanceWidthArray
            b"\x01\xf4"  # .notdef: 500
            b"\x02X"  # space: 600
            b"\x02&"  # A: 550
            b"\x00\x00"  # acutecomb: 0
            b"\x02&"  # Aacute: 550
            b"\x01\xf4"  # zero: 500
        )

        hmtxTable.reconstruct(data, ttFont)

        assert hmtxTable.metrics == {
            ".notdef": (500, 50),
            "space": (600, 0),
            "A": (550, 40),
            "acutecomb": (0, -40),
            "Aacute": (550, 40),
            "zero": (500, 30),
            "one": (500, 50),
            "two": (500, 40),
        }

    def test_reconstruct_proportional_sidebearings(self, ttFont):
        hmtxTable = WOFF2HmtxTable()

        data = (
            b"\x02"  # 00000010 | bits 0 unset: explicit proportional sidebearings
            # advanceWidthArray
            b"\x01\xf4"  # .notdef: 500
            b"\x02X"  # space: 600
            b"\x02&"  # A: 550
            b"\x00\x00"  # acutecomb: 0
            b"\x02&"  # Aacute: 550
            b"\x01\xf4"  # zero: 500
            # lsbArray
            b"\x002"  # .notdef: 50
            b"\x00\x00"  # space: 0
            b"\x00'"  # A: 39 (xMin: 40)
            b"\xff\xd8"  # acutecomb: -40
            b"\x00("  # Aacute: 40
            b"\x00\x1e"  # zero: 30
        )

        hmtxTable.reconstruct(data, ttFont)

        assert hmtxTable.metrics == {
            ".notdef": (500, 50),
            "space": (600, 0),
            "A": (550, 39),
            "acutecomb": (0, -40),
            "Aacute": (550, 40),
            "zero": (500, 30),
            "one": (500, 50),
            "two": (500, 40),
        }

        assert ttFont["glyf"]["A"].xMin == 40

    def test_reconstruct_monospaced_sidebearings(self, ttFont):
        hmtxTable = WOFF2HmtxTable()

        data = (
            b"\x01"  # 00000001 | bits 1 unset: explicit monospaced sidebearings
            # advanceWidthArray
            b"\x01\xf4"  # .notdef: 500
            b"\x02X"  # space: 600
            b"\x02&"  # A: 550
            b"\x00\x00"  # acutecomb: 0
            b"\x02&"  # Aacute: 550
            b"\x01\xf4"  # zero: 500
            # leftSideBearingArray
            b"\x003"  # one: 51 (xMin: 50)
            b"\x00("  # two: 40
        )

        hmtxTable.reconstruct(data, ttFont)

        assert hmtxTable.metrics == {
            ".notdef": (500, 50),
            "space": (600, 0),
            "A": (550, 40),
            "acutecomb": (0, -40),
            "Aacute": (550, 40),
            "zero": (500, 30),
            "one": (500, 51),
            "two": (500, 40),
        }

        assert ttFont["glyf"]["one"].xMin == 50

    def test_reconstruct_flags_reserved_bits(self):
        hmtxTable = WOFF2HmtxTable()

        with pytest.raises(
            ttLib.TTLibError, match="Bits 2-7 of 'hmtx' flags are reserved"
        ):
            hmtxTable.reconstruct(b"\xFF", ttFont=None)

    def test_reconstruct_flags_required_bits(self):
        hmtxTable = WOFF2HmtxTable()

        with pytest.raises(ttLib.TTLibError, match="either bits 0 or 1 .* must set"):
            hmtxTable.reconstruct(b"\x00", ttFont=None)

    def test_reconstruct_too_much_data(self, ttFont):
        ttFont["hhea"].numberOfHMetrics = 2
        data = b"\x03\x01\xf4\x02X\x02&"
        hmtxTable = WOFF2HmtxTable()

        with pytest.raises(ttLib.TTLibError, match="too much 'hmtx' table data"):
            hmtxTable.reconstruct(data, ttFont)


class WOFF2RoundtripTest(object):
    @staticmethod
    def roundtrip(infile):
        infile.seek(0)
        ttFont = ttLib.TTFont(infile, recalcBBoxes=False, recalcTimestamp=False)
        outfile = BytesIO()
        ttFont.save(outfile)
        return outfile, ttFont

    def test_roundtrip_default_transforms(self, ttFont):
        ttFont.flavor = "woff2"
        # ttFont.flavorData = None
        tmp = BytesIO()
        ttFont.save(tmp)

        tmp2, ttFont2 = self.roundtrip(tmp)

        assert tmp.getvalue() == tmp2.getvalue()
        assert ttFont2.reader.flavorData.transformedTables == {"glyf", "loca"}

    def test_roundtrip_no_transforms(self, ttFont):
        ttFont.flavor = "woff2"
        ttFont.flavorData = WOFF2FlavorData(transformedTables=[])
        tmp = BytesIO()
        ttFont.save(tmp)

        tmp2, ttFont2 = self.roundtrip(tmp)

        assert tmp.getvalue() == tmp2.getvalue()
        assert not ttFont2.reader.flavorData.transformedTables

    def test_roundtrip_all_transforms(self, ttFont):
        ttFont.flavor = "woff2"
        ttFont.flavorData = WOFF2FlavorData(transformedTables=["glyf", "loca", "hmtx"])
        tmp = BytesIO()
        ttFont.save(tmp)

        tmp2, ttFont2 = self.roundtrip(tmp)

        assert tmp.getvalue() == tmp2.getvalue()
        assert ttFont2.reader.flavorData.transformedTables == {"glyf", "loca", "hmtx"}

    def test_roundtrip_only_hmtx_no_glyf_transform(self, ttFont):
        ttFont.flavor = "woff2"
        ttFont.flavorData = WOFF2FlavorData(transformedTables=["hmtx"])
        tmp = BytesIO()
        ttFont.save(tmp)

        tmp2, ttFont2 = self.roundtrip(tmp)

        assert tmp.getvalue() == tmp2.getvalue()
        assert ttFont2.reader.flavorData.transformedTables == {"hmtx"}

    def test_roundtrip_no_glyf_and_loca_tables(self):
        ttx = os.path.join(
            os.path.dirname(current_dir), "subset", "data", "google_color.ttx"
        )
        ttFont = ttLib.TTFont()
        ttFont.importXML(ttx)

        assert "glyf" not in ttFont
        assert "loca" not in ttFont

        ttFont.flavor = "woff2"
        tmp = BytesIO()
        ttFont.save(tmp)

        tmp2, ttFont2 = self.roundtrip(tmp)
        assert tmp.getvalue() == tmp2.getvalue()
        assert ttFont.flavor == "woff2"

    def test_roundtrip_off_curve_despite_overlap_bit(self):
        ttx = os.path.join(data_dir, "woff2_overlap_offcurve_in.ttx")
        ttFont = ttLib.TTFont()
        ttFont.importXML(ttx)

        assert ttFont["glyf"]["A"].flags[0] == _g_l_y_f.flagOverlapSimple

        ttFont.flavor = "woff2"
        tmp = BytesIO()
        ttFont.save(tmp)

        _, ttFont2 = self.roundtrip(tmp)
        assert ttFont2.flavor == "woff2"
        # check that the off-curve point is still there
        assert ttFont2["glyf"]["A"].flags[0] & _g_l_y_f.flagOnCurve == 0
        # check that the overlap bit is still there
        assert ttFont2["glyf"]["A"].flags[0] & _g_l_y_f.flagOverlapSimple != 0


class MainTest(object):
    @staticmethod
    def make_ttf(tmpdir):
        ttFont = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
        ttFont.importXML(TTX)
        filename = str(tmpdir / "TestTTF-Regular.ttf")
        ttFont.save(filename)
        return filename

    def test_compress_ttf(self, tmpdir):
        input_file = self.make_ttf(tmpdir)

        assert woff2.main(["compress", input_file]) is None

        assert (tmpdir / "TestTTF-Regular.woff2").check(file=True)

    def test_compress_ttf_no_glyf_transform(self, tmpdir):
        input_file = self.make_ttf(tmpdir)

        assert woff2.main(["compress", "--no-glyf-transform", input_file]) is None

        assert (tmpdir / "TestTTF-Regular.woff2").check(file=True)

    def test_compress_ttf_hmtx_transform(self, tmpdir):
        input_file = self.make_ttf(tmpdir)

        assert woff2.main(["compress", "--hmtx-transform", input_file]) is None

        assert (tmpdir / "TestTTF-Regular.woff2").check(file=True)

    def test_compress_ttf_no_glyf_transform_hmtx_transform(self, tmpdir):
        input_file = self.make_ttf(tmpdir)

        assert (
            woff2.main(
                ["compress", "--no-glyf-transform", "--hmtx-transform", input_file]
            )
            is None
        )

        assert (tmpdir / "TestTTF-Regular.woff2").check(file=True)

    def test_compress_output_file(self, tmpdir):
        input_file = self.make_ttf(tmpdir)
        output_file = tmpdir / "TestTTF.woff2"

        assert woff2.main(["compress", "-o", str(output_file), str(input_file)]) is None

        assert output_file.check(file=True)

    def test_compress_otf(self, tmpdir):
        ttFont = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
        ttFont.importXML(OTX)
        input_file = str(tmpdir / "TestOTF-Regular.otf")
        ttFont.save(input_file)

        assert woff2.main(["compress", input_file]) is None

        assert (tmpdir / "TestOTF-Regular.woff2").check(file=True)

    def test_recompress_woff2_keeps_flavorData(self, tmpdir):
        woff2_font = ttLib.TTFont(BytesIO(TT_WOFF2.getvalue()))
        woff2_font.flavorData.privData = b"FOOBAR"
        woff2_file = tmpdir / "TestTTF-Regular.woff2"
        woff2_font.save(str(woff2_file))

        assert woff2_font.flavorData.transformedTables == {"glyf", "loca"}

        woff2.main(["compress", "--hmtx-transform", str(woff2_file)])

        output_file = tmpdir / "TestTTF-Regular#1.woff2"
        assert output_file.check(file=True)

        new_woff2_font = ttLib.TTFont(str(output_file))

        assert new_woff2_font.flavorData.transformedTables == {"glyf", "loca", "hmtx"}
        assert new_woff2_font.flavorData.privData == b"FOOBAR"

    def test_decompress_ttf(self, tmpdir):
        input_file = tmpdir / "TestTTF-Regular.woff2"
        input_file.write_binary(TT_WOFF2.getvalue())

        assert woff2.main(["decompress", str(input_file)]) is None

        assert (tmpdir / "TestTTF-Regular.ttf").check(file=True)

    def test_decompress_otf(self, tmpdir):
        input_file = tmpdir / "TestTTF-Regular.woff2"
        input_file.write_binary(CFF_WOFF2.getvalue())

        assert woff2.main(["decompress", str(input_file)]) is None

        assert (tmpdir / "TestTTF-Regular.otf").check(file=True)

    def test_decompress_output_file(self, tmpdir):
        input_file = tmpdir / "TestTTF-Regular.woff2"
        input_file.write_binary(TT_WOFF2.getvalue())
        output_file = tmpdir / "TestTTF.ttf"

        assert (
            woff2.main(["decompress", "-o", str(output_file), str(input_file)]) is None
        )

        assert output_file.check(file=True)

    def test_no_subcommand_show_help(self, capsys):
        with pytest.raises(SystemExit):
            woff2.main(["--help"])

        captured = capsys.readouterr()
        assert "usage: fonttools ttLib.woff2" in captured.out


class Base128Test(unittest.TestCase):
    def test_unpackBase128(self):
        self.assertEqual(unpackBase128(b"\x3f\x00\x00"), (63, b"\x00\x00"))
        self.assertEqual(unpackBase128(b"\x8f\xff\xff\xff\x7f")[0], 4294967295)

        self.assertRaisesRegex(
            ttLib.TTLibError,
            "UIntBase128 value must not start with leading zeros",
            unpackBase128,
            b"\x80\x80\x3f",
        )

        self.assertRaisesRegex(
            ttLib.TTLibError,
            "UIntBase128-encoded sequence is longer than 5 bytes",
            unpackBase128,
            b"\x8f\xff\xff\xff\xff\x7f",
        )

        self.assertRaisesRegex(
            ttLib.TTLibError,
            r"UIntBase128 value exceeds 2\*\*32-1",
            unpackBase128,
            b"\x90\x80\x80\x80\x00",
        )

        self.assertRaisesRegex(
            ttLib.TTLibError,
            "not enough data to unpack UIntBase128",
            unpackBase128,
            b"",
        )

    def test_base128Size(self):
        self.assertEqual(base128Size(0), 1)
        self.assertEqual(base128Size(24567), 3)
        self.assertEqual(base128Size(2**32 - 1), 5)

    def test_packBase128(self):
        self.assertEqual(packBase128(63), b"\x3f")
        self.assertEqual(packBase128(2**32 - 1), b"\x8f\xff\xff\xff\x7f")
        self.assertRaisesRegex(
            ttLib.TTLibError,
            r"UIntBase128 format requires 0 <= integer <= 2\*\*32-1",
            packBase128,
            2**32 + 1,
        )
        self.assertRaisesRegex(
            ttLib.TTLibError,
            r"UIntBase128 format requires 0 <= integer <= 2\*\*32-1",
            packBase128,
            -1,
        )


class UShort255Test(unittest.TestCase):
    def test_unpack255UShort(self):
        self.assertEqual(unpack255UShort(bytechr(252))[0], 252)
        # some numbers (e.g. 506) can have multiple encodings
        self.assertEqual(unpack255UShort(struct.pack(b"BB", 254, 0))[0], 506)
        self.assertEqual(unpack255UShort(struct.pack(b"BB", 255, 253))[0], 506)
        self.assertEqual(unpack255UShort(struct.pack(b"BBB", 253, 1, 250))[0], 506)

        self.assertRaisesRegex(
            ttLib.TTLibError,
            "not enough data to unpack 255UInt16",
            unpack255UShort,
            struct.pack(b"BB", 253, 0),
        )

        self.assertRaisesRegex(
            ttLib.TTLibError,
            "not enough data to unpack 255UInt16",
            unpack255UShort,
            struct.pack(b"B", 254),
        )

        self.assertRaisesRegex(
            ttLib.TTLibError,
            "not enough data to unpack 255UInt16",
            unpack255UShort,
            struct.pack(b"B", 255),
        )

    def test_pack255UShort(self):
        self.assertEqual(pack255UShort(252), b"\xfc")
        self.assertEqual(pack255UShort(505), b"\xff\xfc")
        self.assertEqual(pack255UShort(506), b"\xfe\x00")
        self.assertEqual(pack255UShort(762), b"\xfd\x02\xfa")

        self.assertRaisesRegex(
            ttLib.TTLibError,
            "255UInt16 format requires 0 <= integer <= 65535",
            pack255UShort,
            -1,
        )

        self.assertRaisesRegex(
            ttLib.TTLibError,
            "255UInt16 format requires 0 <= integer <= 65535",
            pack255UShort,
            0xFFFF + 1,
        )


class VarCompositeTest(unittest.TestCase):
    def test_var_composite(self):
        input_path = os.path.join(data_dir, "varc-ac00-ac01.ttf")
        ttf = ttLib.TTFont(input_path)
        ttf.flavor = "woff2"
        out = BytesIO()
        ttf.save(out)

        ttf = ttLib.TTFont(out)
        ttf.flavor = None
        out = BytesIO()
        ttf.save(out)


class CubicTest(unittest.TestCase):
    def test_cubic(self):
        input_path = os.path.join(
            data_dir, "..", "tables", "data", "NotoSans-VF-cubic.subset.ttf"
        )
        ttf = ttLib.TTFont(input_path)
        pen1 = RecordingPen()
        ttf.getGlyphSet()["a"].draw(pen1)
        ttf.flavor = "woff2"
        out = BytesIO()
        ttf.save(out)

        ttf = ttLib.TTFont(out)
        ttf.flavor = None
        pen2 = RecordingPen()
        ttf.getGlyphSet()["a"].draw(pen2)
        out = BytesIO()
        ttf.save(out)

        assert pen1.value == pen2.value


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from io import BytesIO
import os
import unittest
from fontTools.misc.textTools import bytesjoin, tobytes
from fontTools.misc.xmlWriter import XMLWriter

HEADER = b'<?xml version="1.0" encoding="UTF-8"?>\n'


class TestXMLWriter(unittest.TestCase):
    def test_comment_escaped(self):
        writer = XMLWriter(BytesIO())
        writer.comment("This&that are <comments>")
        self.assertEqual(
            HEADER + b"<!-- This&amp;that are &lt;comments&gt; -->",
            writer.file.getvalue(),
        )

    def test_comment_multiline(self):
        writer = XMLWriter(BytesIO())
        writer.comment("Hello world\nHow are you?")
        self.assertEqual(
            HEADER + b"<!-- Hello world\n     How are you? -->", writer.file.getvalue()
        )

    def test_encoding_default(self):
        writer = XMLWriter(BytesIO())
        self.assertEqual(
            b'<?xml version="1.0" encoding="UTF-8"?>\n', writer.file.getvalue()
        )

    def test_encoding_utf8(self):
        # https://github.com/fonttools/fonttools/issues/246
        writer = XMLWriter(BytesIO(), encoding="utf8")
        self.assertEqual(
            b'<?xml version="1.0" encoding="UTF-8"?>\n', writer.file.getvalue()
        )

    def test_encoding_UTF_8(self):
        # https://github.com/fonttools/fonttools/issues/246
        writer = XMLWriter(BytesIO(), encoding="UTF-8")
        self.assertEqual(
            b'<?xml version="1.0" encoding="UTF-8"?>\n', writer.file.getvalue()
        )

    def test_encoding_UTF8(self):
        # https://github.com/fonttools/fonttools/issues/246
        writer = XMLWriter(BytesIO(), encoding="UTF8")
        self.assertEqual(
            b'<?xml version="1.0" encoding="UTF-8"?>\n', writer.file.getvalue()
        )

    def test_encoding_other(self):
        self.assertRaises(Exception, XMLWriter, BytesIO(), encoding="iso-8859-1")

    def test_write(self):
        writer = XMLWriter(BytesIO())
        writer.write("foo&bar")
        self.assertEqual(HEADER + b"foo&amp;bar", writer.file.getvalue())

    def test_indent_dedent(self):
        writer = XMLWriter(BytesIO())
        writer.write("foo")
        writer.newline()
        writer.indent()
        writer.write("bar")
        writer.newline()
        writer.dedent()
        writer.write("baz")
        self.assertEqual(
            HEADER + bytesjoin(["foo", "  bar", "baz"], "\n"), writer.file.getvalue()
        )

    def test_writecdata(self):
        writer = XMLWriter(BytesIO())
        writer.writecdata("foo&bar")
        self.assertEqual(HEADER + b"<![CDATA[foo&bar]]>", writer.file.getvalue())

    def test_simpletag(self):
        writer = XMLWriter(BytesIO())
        writer.simpletag("tag", a="1", b="2")
        self.assertEqual(HEADER + b'<tag a="1" b="2"/>', writer.file.getvalue())

    def test_begintag_endtag(self):
        writer = XMLWriter(BytesIO())
        writer.begintag("tag", attr="value")
        writer.write("content")
        writer.endtag("tag")
        self.assertEqual(
            HEADER + b'<tag attr="value">content</tag>', writer.file.getvalue()
        )

    def test_dumphex(self):
        writer = XMLWriter(BytesIO())
        writer.dumphex(
            "Type is a beautiful group of letters, not a group of beautiful letters."
        )
        self.assertEqual(
            HEADER
            + bytesjoin(
                [
                    "54797065 20697320 61206265 61757469",
                    "66756c20 67726f75 70206f66 206c6574",
                    "74657273 2c206e6f 74206120 67726f75",
                    "70206f66 20626561 75746966 756c206c",
                    "65747465 72732e  ",
                    "",
                ],
                joiner="\n",
            ),
            writer.file.getvalue(),
        )

    def test_stringifyattrs(self):
        writer = XMLWriter(BytesIO())
        expected = ' attr="0"'
        self.assertEqual(expected, writer.stringifyattrs(attr=0))
        self.assertEqual(expected, writer.stringifyattrs(attr=b"0"))
        self.assertEqual(expected, writer.stringifyattrs(attr="0"))
        self.assertEqual(expected, writer.stringifyattrs(attr="0"))

    def test_carriage_return_escaped(self):
        writer = XMLWriter(BytesIO())
        writer.write("two lines\r\nseparated by Windows line endings")
        self.assertEqual(
            HEADER + b"two lines&#13;\nseparated by Windows line endings",
            writer.file.getvalue(),
        )

    def test_newlinestr(self):
        header = b'<?xml version="1.0" encoding="UTF-8"?>'

        for nls in (None, "\n", "\r\n", "\r", ""):
            writer = XMLWriter(BytesIO(), newlinestr=nls)
            writer.write("hello")
            writer.newline()
            writer.write("world")
            writer.newline()

            linesep = tobytes(os.linesep) if nls is None else tobytes(nls)

            self.assertEqual(
                header + linesep + b"hello" + linesep + b"world" + linesep,
                writer.file.getvalue(),
            )


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



import glob
import os

import pytest

from fontTools import tfmLib

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


@pytest.mark.parametrize("path", glob.glob(f"{DATA_DIR}/cm*.tfm"))
def test_read(path):
    tfm = tfmLib.TFM(path)
    assert tfm.designsize == 10.0
    assert tfm.fontdimens
    assert len(tfm.fontdimens) >= 7
    assert tfm.extraheader == {}
    assert tfm.right_boundary_char is None
    assert tfm.left_boundary_char is None
    assert len(tfm.chars) == 128


def test_read_boundary_char():
    path = os.path.join(DATA_DIR, "dummy-space.tfm")
    tfm = tfmLib.TFM(path)
    assert tfm.right_boundary_char == 1
    assert tfm.left_boundary_char == 256


def test_read_fontdimens_vanilla():
    path = os.path.join(DATA_DIR, "cmr10.tfm")
    tfm = tfmLib.TFM(path)
    assert tfm.fontdimens == {
        "SLANT": 0.0,
        "SPACE": 0.33333396911621094,
        "STRETCH": 0.16666698455810547,
        "SHRINK": 0.11111164093017578,
        "XHEIGHT": 0.4305553436279297,
        "QUAD": 1.0000028610229492,
        "EXTRASPACE": 0.11111164093017578,
    }


def test_read_fontdimens_mathex():
    path = os.path.join(DATA_DIR, "cmex10.tfm")
    tfm = tfmLib.TFM(path)
    assert tfm.fontdimens == {
        "SLANT": 0.0,
        "SPACE": 0.0,
        "STRETCH": 0.0,
        "SHRINK": 0.0,
        "XHEIGHT": 0.4305553436279297,
        "QUAD": 1.0000028610229492,
        "EXTRASPACE": 0.0,
        "DEFAULTRULETHICKNESS": 0.03999900817871094,
        "BIGOPSPACING1": 0.11111164093017578,
        "BIGOPSPACING2": 0.16666698455810547,
        "BIGOPSPACING3": 0.19999980926513672,
        "BIGOPSPACING4": 0.6000003814697266,
        "BIGOPSPACING5": 0.10000038146972656,
    }


def test_read_fontdimens_mathsy():
    path = os.path.join(DATA_DIR, "cmsy10.tfm")
    tfm = tfmLib.TFM(path)
    assert tfm.fontdimens == {
        "SLANT": 0.25,
        "SPACE": 0.0,
        "STRETCH": 0.0,
        "SHRINK": 0.0,
        "XHEIGHT": 0.4305553436279297,
        "QUAD": 1.0000028610229492,
        "EXTRASPACE": 0.0,
        "NUM1": 0.6765079498291016,
        "NUM2": 0.39373207092285156,
        "NUM3": 0.44373130798339844,
        "DENOM1": 0.6859512329101562,
        "DENOM2": 0.34484100341796875,
        "SUP1": 0.41289234161376953,
        "SUP2": 0.36289215087890625,
        "SUP3": 0.28888893127441406,
        "SUB1": 0.14999961853027344,
        "SUB2": 0.24721717834472656,
        "SUBDROP": 0.05000019073486328,
        "SUPDROP": 0.3861083984375,
        "DELIM1": 2.3899993896484375,
        "DELIM2": 1.010000228881836,
        "AXISHEIGHT": 0.25,
    }



import os
from fontTools import ttLib
import pytest


TESTDATA = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture
def varfont():
    f = ttLib.TTFont()
    f.importXML(os.path.join(TESTDATA, "PartialInstancerTest-VF.ttx"))
    return f



from fontTools.misc.testTools import getXML, parseXML
from fontTools.misc.textTools import deHexStr
from fontTools.ttLib import getTableModule, newTable
import unittest


CPAL_DATA_V0 = deHexStr(
    "0000 0002 "  # version=0, numPaletteEntries=2
    "0002 0004 "  # numPalettes=2, numColorRecords=4
    "00000010 "  # offsetToFirstColorRecord=16
    "0000 0002 "  # colorRecordIndex=[0, 2]
    "000000FF FFCC66FF "  # colorRecord #0, #1 (blue/green/red/alpha)
    "000000FF 000080FF"
)  # colorRecord #2, #3


CPAL_DATA_V0_SHARING_COLORS = deHexStr(
    "0000 0003 "  # version=0, numPaletteEntries=3
    "0004 0006 "  # numPalettes=4, numColorRecords=6
    "00000014 "  # offsetToFirstColorRecord=20
    "0000 0000 0003 0000 "  # colorRecordIndex=[0, 0, 3, 0]
    "443322FF 77889911 55555555 "  # colorRecord #0, #1, #2 (BGRA)
    "443322FF 77889911 FFFFFFFF"
)  # colorRecord #3, #4, #5


CPAL_DATA_V1_NOLABELS_NOTYPES = deHexStr(
    "0001 0003 "  # version=1, numPaletteEntries=3
    "0002 0006 "  # numPalettes=2, numColorRecords=6
    "0000001C  "  # offsetToFirstColorRecord=28
    "0000 0003 "  # colorRecordIndex=[0, 3]
    "00000000  "  # offsetToPaletteTypeArray=0
    "00000000  "  # offsetToPaletteLabelArray=0
    "00000000  "  # offsetToPaletteEntryLabelArray=0
    "CAFECAFE 00112233 44556677 "  # colorRecord #0, #1, #2 (BGRA)
    "31415927 42424242 00331337"
)  # colorRecord #3, #4, #5


CPAL_DATA_V1 = deHexStr(
    "0001 0003 "  # version=1, numPaletteEntries=3
    "0002 0006 "  # numPalettes=2, numColorRecords=6
    "0000001C  "  # offsetToFirstColorRecord=28
    "0000 0003 "  # colorRecordIndex=[0, 3]
    "00000034  "  # offsetToPaletteTypeArray=52
    "0000003C  "  # offsetToPaletteLabelArray=60
    "00000040  "  # offsetToPaletteEntryLabelArray=64
    "CAFECAFE 00112233 44556677 "  # colorRecord #0, #1, #2 (BGRA)
    "31415927 42424242 00331337 "  # colorRecord #3, #4, #5
    "00000001 00000002 "  # paletteType=[1, 2]
    "0102 0103 "  # paletteLabel=[258, 259]
    "0201 0202 0203"
)  # paletteEntryLabel=[513, 514, 515]


class FakeNameTable(object):
    def __init__(self, names):
        self.names = names

    def getDebugName(self, nameID):
        return self.names.get(nameID)


class CPALTest(unittest.TestCase):
    def test_decompile_v0(self):
        cpal = newTable("CPAL")
        cpal.decompile(CPAL_DATA_V0, ttFont=None)
        self.assertEqual(cpal.version, 0)
        self.assertEqual(cpal.numPaletteEntries, 2)
        self.assertEqual(
            repr(cpal.palettes), "[[#000000FF, #66CCFFFF], [#000000FF, #800000FF]]"
        )

    def test_decompile_v0_sharingColors(self):
        cpal = newTable("CPAL")
        cpal.decompile(CPAL_DATA_V0_SHARING_COLORS, ttFont=None)
        self.assertEqual(cpal.version, 0)
        self.assertEqual(cpal.numPaletteEntries, 3)
        self.assertEqual(
            [repr(p) for p in cpal.palettes],
            [
                "[#223344FF, #99887711, #55555555]",
                "[#223344FF, #99887711, #55555555]",
                "[#223344FF, #99887711, #FFFFFFFF]",
                "[#223344FF, #99887711, #55555555]",
            ],
        )

    def test_decompile_v1_noLabelsNoTypes(self):
        cpal = newTable("CPAL")
        cpal.decompile(CPAL_DATA_V1_NOLABELS_NOTYPES, ttFont=None)
        self.assertEqual(cpal.version, 1)
        self.assertEqual(cpal.numPaletteEntries, 3)
        self.assertEqual(
            [repr(p) for p in cpal.palettes],
            [
                "[#CAFECAFE, #22110033, #66554477]",  # RGBA
                "[#59413127, #42424242, #13330037]",
            ],
        )
        self.assertEqual(cpal.paletteLabels, [cpal.NO_NAME_ID] * len(cpal.palettes))
        self.assertEqual(cpal.paletteTypes, [0, 0])
        self.assertEqual(
            cpal.paletteEntryLabels, [cpal.NO_NAME_ID] * cpal.numPaletteEntries
        )

    def test_decompile_v1(self):
        cpal = newTable("CPAL")
        cpal.decompile(CPAL_DATA_V1, ttFont=None)
        self.assertEqual(cpal.version, 1)
        self.assertEqual(cpal.numPaletteEntries, 3)
        self.assertEqual(
            [repr(p) for p in cpal.palettes],
            [
                "[#CAFECAFE, #22110033, #66554477]",  # RGBA
                "[#59413127, #42424242, #13330037]",
            ],
        )
        self.assertEqual(cpal.paletteTypes, [1, 2])
        self.assertEqual(cpal.paletteLabels, [258, 259])
        self.assertEqual(cpal.paletteEntryLabels, [513, 514, 515])

    def test_compile_v0(self):
        cpal = newTable("CPAL")
        cpal.decompile(CPAL_DATA_V0, ttFont=None)
        self.assertEqual(cpal.compile(ttFont=None), CPAL_DATA_V0)

    def test_compile_v0_sharingColors(self):
        cpal = newTable("CPAL")
        cpal.version = 0
        Color = getTableModule("CPAL").Color
        palette1 = [
            Color(red=0x22, green=0x33, blue=0x44, alpha=0xFF),
            Color(red=0x99, green=0x88, blue=0x77, alpha=0x11),
            Color(red=0x55, green=0x55, blue=0x55, alpha=0x55),
        ]
        palette2 = [
            Color(red=0x22, green=0x33, blue=0x44, alpha=0xFF),
            Color(red=0x99, green=0x88, blue=0x77, alpha=0x11),
            Color(red=0xFF, green=0xFF, blue=0xFF, alpha=0xFF),
        ]
        cpal.numPaletteEntries = len(palette1)
        cpal.palettes = [palette1, palette1, palette2, palette1]
        self.assertEqual(cpal.compile(ttFont=None), CPAL_DATA_V0_SHARING_COLORS)

    def test_compile_v1(self):
        cpal = newTable("CPAL")
        cpal.decompile(CPAL_DATA_V1, ttFont=None)
        self.assertEqual(cpal.compile(ttFont=None), CPAL_DATA_V1)

    def test_compile_v1_noLabelsNoTypes(self):
        cpal = newTable("CPAL")
        cpal.decompile(CPAL_DATA_V1_NOLABELS_NOTYPES, ttFont=None)
        self.assertEqual(cpal.compile(ttFont=None), CPAL_DATA_V1_NOLABELS_NOTYPES)

    def test_toXML_v0(self):
        cpal = newTable("CPAL")
        cpal.decompile(CPAL_DATA_V0, ttFont=None)
        self.assertEqual(
            getXML(cpal.toXML),
            [
                '<version value="0"/>',
                '<numPaletteEntries value="2"/>',
                '<palette index="0">',
                '  <color index="0" value="#000000FF"/>',
                '  <color index="1" value="#66CCFFFF"/>',
                "</palette>",
                '<palette index="1">',
                '  <color index="0" value="#000000FF"/>',
                '  <color index="1" value="#800000FF"/>',
                "</palette>",
            ],
        )

    def test_toXML_v1(self):
        name = FakeNameTable(
            {258: "Spring theme", 259: "Winter theme", 513: "darks", 515: "lights"}
        )
        cpal = newTable("CPAL")
        ttFont = {"name": name, "CPAL": cpal}
        cpal.decompile(CPAL_DATA_V1, ttFont)
        self.assertEqual(
            getXML(cpal.toXML, ttFont),
            [
                '<version value="1"/>',
                '<numPaletteEntries value="3"/>',
                '<palette index="0" label="258" type="1">',
                "  <!-- Spring theme -->",
                '  <color index="0" value="#CAFECAFE"/>',
                '  <color index="1" value="#22110033"/>',
                '  <color index="2" value="#66554477"/>',
                "</palette>",
                '<palette index="1" label="259" type="2">',
                "  <!-- Winter theme -->",
                '  <color index="0" value="#59413127"/>',
                '  <color index="1" value="#42424242"/>',
                '  <color index="2" value="#13330037"/>',
                "</palette>",
                "<paletteEntryLabels>",
                '  <label index="0" value="513"/><!-- darks -->',
                '  <label index="1" value="514"/>',
                '  <label index="2" value="515"/><!-- lights -->',
                "</paletteEntryLabels>",
            ],
        )

    def test_fromXML_v0(self):
        cpal = newTable("CPAL")
        for name, attrs, content in parseXML(
            '<version value="0"/>'
            '<numPaletteEntries value="2"/>'
            '<palette index="0">'
            '  <color index="0" value="#12345678"/>'
            '  <color index="1" value="#FEDCBA98"/>'
            "</palette>"
        ):
            cpal.fromXML(name, attrs, content, ttFont=None)
        self.assertEqual(cpal.version, 0)
        self.assertEqual(cpal.numPaletteEntries, 2)
        self.assertEqual(repr(cpal.palettes), "[[#12345678, #FEDCBA98]]")

    def test_fromXML_v1(self):
        cpal = newTable("CPAL")
        for name, attrs, content in parseXML(
            '<version value="1"/>'
            '<numPaletteEntries value="3"/>'
            '<palette index="0" label="259" type="2">'
            '  <color index="0" value="#12345678"/>'
            '  <color index="1" value="#FEDCBA98"/>'
            '  <color index="2" value="#CAFECAFE"/>'
            "</palette>"
            "<paletteEntryLabels>"
            '  <label index="1" value="262"/>'
            "</paletteEntryLabels>"
        ):
            cpal.fromXML(name, attrs, content, ttFont=None)
        self.assertEqual(cpal.version, 1)
        self.assertEqual(cpal.numPaletteEntries, 3)
        self.assertEqual(repr(cpal.palettes), "[[#12345678, #FEDCBA98, #CAFECAFE]]")
        self.assertEqual(cpal.paletteLabels, [259])
        self.assertEqual(cpal.paletteTypes, [2])
        self.assertEqual(
            cpal.paletteEntryLabels, [cpal.NO_NAME_ID, 262, cpal.NO_NAME_ID]
        )


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from fontTools.misc.testTools import parseXML
from fontTools.misc.textTools import deHexStr
from fontTools.misc.xmlWriter import XMLWriter
from fontTools.ttLib import TTLibError
from fontTools.ttLib.tables._m_e_t_a import table__m_e_t_a
from io import BytesIO
import unittest


# From a real font on MacOS X, but substituted 'bild' tag by 'TEST',
# and shortened the payload.
META_DATA = deHexStr(
    "00 00 00 01 00 00 00 00 00 00 00 1C 00 00 00 01 "
    "54 45 53 54 00 00 00 1C 00 00 00 04 CA FE BE EF"
)

# The 'dlng' and 'slng' tag with text data containing "augmented" BCP 47
# comma-separated or comma-space-separated tags. These should be UTF-8 encoded
# text.
META_DATA_TEXT = deHexStr(
    "00 00 00 01 00 00 00 00 00 00 00 28 00 00 00 02 "
    "64 6C 6E 67 00 00 00 28 00 00 00 0E 73 6C 6E 67 "
    "00 00 00 36 00 00 00 0E 4C 61 74 6E 2C 47 72 65 "
    "6B 2C 43 79 72 6C 4C 61 74 6E 2C 47 72 65 6B 2C "
    "43 79 72 6C"
)


class MetaTableTest(unittest.TestCase):
    def test_decompile(self):
        table = table__m_e_t_a()
        table.decompile(META_DATA, ttFont={"meta": table})
        self.assertEqual({"TEST": b"\xCA\xFE\xBE\xEF"}, table.data)

    def test_compile(self):
        table = table__m_e_t_a()
        table.data["TEST"] = b"\xCA\xFE\xBE\xEF"
        self.assertEqual(META_DATA, table.compile(ttFont={"meta": table}))

    def test_decompile_text(self):
        table = table__m_e_t_a()
        table.decompile(META_DATA_TEXT, ttFont={"meta": table})
        self.assertEqual(
            {"dlng": "Latn,Grek,Cyrl", "slng": "Latn,Grek,Cyrl"}, table.data
        )

    def test_compile_text(self):
        table = table__m_e_t_a()
        table.data["dlng"] = "Latn,Grek,Cyrl"
        table.data["slng"] = "Latn,Grek,Cyrl"
        self.assertEqual(META_DATA_TEXT, table.compile(ttFont={"meta": table}))

    def test_toXML(self):
        table = table__m_e_t_a()
        table.data["TEST"] = b"\xCA\xFE\xBE\xEF"
        writer = XMLWriter(BytesIO())
        table.toXML(writer, {"meta": table})
        xml = writer.file.getvalue().decode("utf-8")
        self.assertEqual(
            ['<hexdata tag="TEST">', "cafebeef", "</hexdata>"],
            [line.strip() for line in xml.splitlines()][1:],
        )

    def test_toXML_ascii_data(self):
        table = table__m_e_t_a()
        table.data["TEST"] = b"Hello!"
        writer = XMLWriter(BytesIO())
        table.toXML(writer, {"meta": table})
        xml = writer.file.getvalue().decode("utf-8")
        self.assertEqual(
            [
                '<hexdata tag="TEST">',
                "<!-- ascii: Hello! -->",
                "48656c6c 6f21",
                "</hexdata>",
            ],
            [line.strip() for line in xml.splitlines()][1:],
        )

    def test_fromXML(self):
        table = table__m_e_t_a()
        for name, attrs, content in parseXML(
            '<hexdata tag="TEST">' "    cafebeef" "</hexdata>"
        ):
            table.fromXML(name, attrs, content, ttFont=None)
        self.assertEqual({"TEST": b"\xCA\xFE\xBE\xEF"}, table.data)

    def test_toXML_text(self):
        table = table__m_e_t_a()
        table.data["dlng"] = "Latn,Grek,Cyrl"
        writer = XMLWriter(BytesIO())
        table.toXML(writer, {"meta": table})
        xml = writer.file.getvalue().decode("utf-8")
        self.assertEqual(
            ['<text tag="dlng">', "Latn,Grek,Cyrl", "</text>"],
            [line.strip() for line in xml.splitlines()][1:],
        )

    def test_fromXML_text(self):
        table = table__m_e_t_a()
        for name, attrs, content in parseXML(
            '<text tag="dlng">' "    Latn,Grek,Cyrl" "</text>"
        ):
            table.fromXML(name, attrs, content, ttFont=None)
        self.assertEqual({"dlng": "Latn,Grek,Cyrl"}, table.data)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



#! /usr/bin/env python3

from fontTools.ttLib import TTFont
from fontTools.ttLib.tables import otTables
import sys

if len(sys.argv) != 2:
    print("usage: layout-features.py fontfile.ttf")
    sys.exit(1)
fontfile = sys.argv[1]
if fontfile.rsplit(".", 1)[-1] == "ttx":
    font = TTFont()
    font.importXML(fontfile)
else:
    font = TTFont(fontfile)

for tag in ("GSUB", "GPOS"):
    if not tag in font:
        continue
    print("Table:", tag)
    table = font[tag].table
    if not table.ScriptList or not table.FeatureList:
        continue
    featureRecords = table.FeatureList.FeatureRecord
    for script in table.ScriptList.ScriptRecord:
        print("  Script:", script.ScriptTag)
        if not script.Script:
            print("    Null script.")
            continue
        languages = list(script.Script.LangSysRecord)
        if script.Script.DefaultLangSys:
            defaultlangsys = otTables.LangSysRecord()
            defaultlangsys.LangSysTag = "default"
            defaultlangsys.LangSys = script.Script.DefaultLangSys
            languages.insert(0, defaultlangsys)
        for langsys in languages:
            print("    Language:", langsys.LangSysTag)
            if not langsys.LangSys:
                print("    Null language.")
                continue
            features = [featureRecords[index] for index in langsys.LangSys.FeatureIndex]
            if langsys.LangSys.ReqFeatureIndex != 0xFFFF:
                record = featureRecords[langsys.LangSys.ReqFeatureIndex]
                requiredfeature = otTables.FeatureRecord()
                requiredfeature.FeatureTag = "required(%s)" % record.FeatureTag
                requiredfeature.Feature = record.Feature
                features.insert(0, requiredfeature)
            for feature in features:
                print("      Feature:", feature.FeatureTag)
                lookups = feature.Feature.LookupListIndex
                print("        Lookups:", ",".join(str(l) for l in lookups))



from fontTools.ttLib.tables import otTables  # trigger setup to occur
from fontTools.ttLib.tables.otConverters import UShort
from fontTools.colorLib.table_builder import TableBuilder
import pytest


class WriteMe:
    value = None


def test_intValue_otRound():
    dest = WriteMe()
    converter = UShort("value", None, None)
    TableBuilder()._convert(dest, "value", converter, 85.6)
    assert dest.value == 86, "Should have used otRound"



from fontTools.cffLib import TopDict, PrivateDict, CharStrings
from fontTools.misc.testTools import parseXML, DataFilesHandler
from fontTools.ttLib import TTFont
import copy
import os
import sys
import unittest
from io import BytesIO


class CffLibTest(DataFilesHandler):
    def test_topDict_recalcFontBBox(self):
        topDict = TopDict()
        topDict.CharStrings = CharStrings(None, None, None, PrivateDict(), None, None)
        topDict.CharStrings.fromXML(
            None,
            None,
            parseXML(
                """
            <CharString name=".notdef">
              endchar
            </CharString>
            <CharString name="foo"><!-- [100, -100, 300, 100] -->
              100 -100 rmoveto 200 hlineto 200 vlineto -200 hlineto endchar
            </CharString>
            <CharString name="bar"><!-- [0, 0, 200, 200] -->
              0 0 rmoveto 200 hlineto 200 vlineto -200 hlineto endchar
            </CharString>
            <CharString name="baz"><!-- [-55.1, -55.1, 55.1, 55.1] -->
              -55.1 -55.1 rmoveto 110.2 hlineto 110.2 vlineto -110.2 hlineto endchar
            </CharString>
        """
            ),
        )

        topDict.recalcFontBBox()
        self.assertEqual(topDict.FontBBox, [-56, -100, 300, 200])

    def test_topDict_recalcFontBBox_empty(self):
        topDict = TopDict()
        topDict.CharStrings = CharStrings(None, None, None, PrivateDict(), None, None)
        topDict.CharStrings.fromXML(
            None,
            None,
            parseXML(
                """
            <CharString name=".notdef">
              endchar
            </CharString>
            <CharString name="space">
              123 endchar
            </CharString>
        """
            ),
        )

        topDict.recalcFontBBox()
        self.assertEqual(topDict.FontBBox, [0, 0, 0, 0])

    def test_topDict_set_Encoding(self):
        ttx_path = self.getpath("TestOTF.ttx")
        font = TTFont(recalcBBoxes=False, recalcTimestamp=False)
        font.importXML(ttx_path)

        topDict = font["CFF "].cff.topDictIndex[0]
        encoding = [".notdef"] * 256
        encoding[0x20] = "space"
        topDict.Encoding = encoding

        self.temp_dir()
        save_path = os.path.join(self.tempdir, "TestOTF.otf")
        font.save(save_path)

        font2 = TTFont(save_path)
        topDict2 = font2["CFF "].cff.topDictIndex[0]
        self.assertEqual(topDict2.Encoding[32], "space")

    def test_CFF_deepcopy(self):
        """Test that deepcopying a TTFont with a CFF table does not recurse
        infinitely."""
        ttx_path = os.path.join(
            os.path.dirname(__file__),
            "..",
            "varLib",
            "data",
            "master_ttx_interpolatable_otf",
            "TestFamily2-Master0.ttx",
        )
        font = TTFont(recalcBBoxes=False, recalcTimestamp=False)
        font.importXML(ttx_path)
        copy.deepcopy(font)

    def test_FDSelect_format_4(self):
        ttx_path = self.getpath("TestFDSelect4.ttx")
        font = TTFont(recalcBBoxes=False, recalcTimestamp=False)
        font.importXML(ttx_path)

        self.temp_dir()
        save_path = os.path.join(self.tempdir, "TestOTF.otf")
        font.save(save_path)

        font2 = TTFont(save_path)
        topDict2 = font2["CFF2"].cff.topDictIndex[0]
        self.assertEqual(topDict2.FDSelect.format, 4)
        self.assertEqual(topDict2.FDSelect.gidArray, [0, 0, 1])

    def test_unique_glyph_names(self):
        font_path = self.getpath("LinLibertine_RBI.otf")
        font = TTFont(font_path, recalcBBoxes=False, recalcTimestamp=False)

        glyphOrder = font.getGlyphOrder()
        self.assertEqual(len(glyphOrder), len(set(glyphOrder)))

        self.temp_dir()
        save_path = os.path.join(self.tempdir, "TestOTF.otf")
        font.save(save_path)

        font2 = TTFont(save_path)
        glyphOrder = font2.getGlyphOrder()
        self.assertEqual(len(glyphOrder), len(set(glyphOrder)))


class CFFToCFF2Test(DataFilesHandler):

    def test_conversion(self):
        font_path = self.getpath("CFFToCFF2-1.otf")
        font = TTFont(font_path)
        from fontTools.cffLib.CFFToCFF2 import convertCFFToCFF2

        convertCFFToCFF2(font)
        f = BytesIO()
        font.save(f)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Modified from https://github.com/adobe-type-tools/psautohint/blob/08b346865710ed3c172f1eb581d6ef243b203f99/python/psautohint/ufoFont.py#L800-L838
import hashlib

from fontTools.pens.basePen import MissingComponentError
from fontTools.pens.pointPen import AbstractPointPen


class HashPointPen(AbstractPointPen):
    """
    This pen can be used to check if a glyph's contents (outlines plus
    components) have changed.

    Components are added as the original outline plus each composite's
    transformation.

    Example: You have some TrueType hinting code for a glyph which you want to
    compile. The hinting code specifies a hash value computed with HashPointPen
    that was valid for the glyph's outlines at the time the hinting code was
    written. Now you can calculate the hash for the glyph's current outlines to
    check if the outlines have changed, which would probably make the hinting
    code invalid.

    > glyph = ufo[name]
    > hash_pen = HashPointPen(glyph.width, ufo)
    > glyph.drawPoints(hash_pen)
    > ttdata = glyph.lib.get("public.truetype.instructions", None)
    > stored_hash = ttdata.get("id", None)  # The hash is stored in the "id" key
    > if stored_hash is None or stored_hash != hash_pen.hash:
    >    logger.error(f"Glyph hash mismatch, glyph '{name}' will have no instructions in font.")
    > else:
    >    # The hash values are identical, the outline has not changed.
    >    # Compile the hinting code ...
    >    pass

    If you want to compare a glyph from a source format which supports floating point
    coordinates and transformations against a glyph from a format which has restrictions
    on the precision of floats, e.g. UFO vs. TTF, you must use an appropriate rounding
    function to make the values comparable. For TTF fonts with composites, this
    construct can be used to make the transform values conform to F2Dot14:

    > ttf_hash_pen = HashPointPen(ttf_glyph_width, ttFont.getGlyphSet())
    > ttf_round_pen = RoundingPointPen(ttf_hash_pen, transformRoundFunc=partial(floatToFixedToFloat, precisionBits=14))
    > ufo_hash_pen = HashPointPen(ufo_glyph.width, ufo)
    > ttf_glyph.drawPoints(ttf_round_pen, ttFont["glyf"])
    > ufo_round_pen = RoundingPointPen(ufo_hash_pen, transformRoundFunc=partial(floatToFixedToFloat, precisionBits=14))
    > ufo_glyph.drawPoints(ufo_round_pen)
    > assert ttf_hash_pen.hash == ufo_hash_pen.hash
    """

    def __init__(self, glyphWidth=0, glyphSet=None):
        self.glyphset = glyphSet
        self.data = ["w%s" % round(glyphWidth, 9)]

    @property
    def hash(self):
        data = "".join(self.data)
        if len(data) >= 128:
            data = hashlib.sha512(data.encode("ascii")).hexdigest()
        return data

    def beginPath(self, identifier=None, **kwargs):
        pass

    def endPath(self):
        self.data.append("|")

    def addPoint(
        self,
        pt,
        segmentType=None,
        smooth=False,
        name=None,
        identifier=None,
        **kwargs,
    ):
        if segmentType is None:
            pt_type = "o"  # offcurve
        else:
            pt_type = segmentType[0]
        self.data.append(f"{pt_type}{pt[0]:g}{pt[1]:+g}")

    def addComponent(self, baseGlyphName, transformation, identifier=None, **kwargs):
        tr = "".join([f"{t:+}" for t in transformation])
        self.data.append("[")
        try:
            self.glyphset[baseGlyphName].drawPoints(self)
        except KeyError:
            raise MissingComponentError(baseGlyphName)
        self.data.append(f"({tr})]")



"""Draw statistical shape of a glyph as an ellipse."""

from fontTools.ttLib import TTFont
from fontTools.pens.recordingPen import RecordingPen
from fontTools.pens.cairoPen import CairoPen
from fontTools.pens.statisticsPen import StatisticsPen
import cairo
import math
import sys


font = TTFont(sys.argv[1])
unicode = sys.argv[2]

cmap = font["cmap"].getBestCmap()
gid = cmap[ord(unicode)]

hhea = font["hhea"]
glyphset = font.getGlyphSet()
with cairo.SVGSurface(
    "example.svg", hhea.advanceWidthMax, hhea.ascent - hhea.descent
) as surface:
    context = cairo.Context(surface)
    context.translate(0, +font["hhea"].ascent)
    context.scale(1, -1)

    glyph = glyphset[gid]

    recording = RecordingPen()
    glyph.draw(recording)

    context.translate((hhea.advanceWidthMax - glyph.width) * 0.5, 0)

    pen = CairoPen(glyphset, context)
    glyph.draw(pen)
    context.fill()

    stats = StatisticsPen(glyphset)
    glyph.draw(stats)

    # https://cookierobotics.com/007/
    a = stats.varianceX
    b = stats.covariance
    c = stats.varianceY
    delta = (((a - c) * 0.5) ** 2 + b * b) ** 0.5
    lambda1 = (a + c) * 0.5 + delta  # Major eigenvalue
    lambda2 = (a + c) * 0.5 - delta  # Minor eigenvalue
    theta = math.atan2(lambda1 - a, b) if b != 0 else (math.pi * 0.5 if a < c else 0)
    mult = 4  # Empirical by drawing '.'
    transform = cairo.Matrix()
    transform.translate(stats.meanX, stats.meanY)
    transform.rotate(theta)
    transform.scale(math.sqrt(lambda1), math.sqrt(lambda2))
    transform.scale(mult, mult)

    ellipse_area = math.sqrt(lambda1) * math.sqrt(lambda2) * math.pi / 4 * mult * mult

    if stats.area:
        context.save()
        context.set_line_cap(cairo.LINE_CAP_ROUND)
        context.transform(transform)
        context.move_to(0, 0)
        context.line_to(0, 0)
        context.set_line_width(1)
        context.set_source_rgba(1, 0, 0, abs(stats.area / ellipse_area))
        context.stroke()
        context.restore()

        context.save()
        context.set_line_cap(cairo.LINE_CAP_ROUND)
        context.set_source_rgb(0.8, 0, 0)
        context.translate(stats.meanX, stats.meanY)

        context.move_to(0, 0)
        context.line_to(0, 0)
        context.set_line_width(15)
        context.stroke()

        context.transform(cairo.Matrix(1, 0, stats.slant, 1, 0, 0))
        context.move_to(0, -stats.meanY + font["hhea"].ascent)
        context.line_to(0, -stats.meanY + font["hhea"].descent)
        context.set_line_width(5)
        context.stroke()

        context.restore()



from io import BytesIO
import os
import unittest
from fontTools.ttLib import TTFont
from fontTools.misc.textTools import strjoin
from fontTools.misc.xmlReader import XMLReader, ProgressPrinter, BUFSIZE
import tempfile


class TestXMLReader(unittest.TestCase):
    def test_decode_utf8(self):
        class DebugXMLReader(XMLReader):
            def __init__(self, fileOrPath, ttFont, progress=None):
                super(DebugXMLReader, self).__init__(fileOrPath, ttFont, progress)
                self.contents = []

            def _endElementHandler(self, name):
                if self.stackSize == 3:
                    name, attrs, content = self.root
                    self.contents.append(content)
                super(DebugXMLReader, self)._endElementHandler(name)

        expected = "fôôbär"
        data = (
            """\
<?xml version="1.0" encoding="UTF-8"?>
<ttFont>
  <name>
    <namerecord nameID="1" platformID="3" platEncID="1" langID="0x409">
      %s
    </namerecord>
  </name>
</ttFont>
"""
            % expected
        )

        with BytesIO(data.encode("utf-8")) as tmp:
            reader = DebugXMLReader(tmp, TTFont())
            reader.read()
        content = strjoin(reader.contents[0]).strip()
        self.assertEqual(expected, content)

    def test_normalise_newlines(self):
        class DebugXMLReader(XMLReader):
            def __init__(self, fileOrPath, ttFont, progress=None):
                super(DebugXMLReader, self).__init__(fileOrPath, ttFont, progress)
                self.newlines = []

            def _characterDataHandler(self, data):
                self.newlines.extend([c for c in data if c in ("\r", "\n")])

        # notice how when CR is escaped, it is not normalised by the XML parser
        data = (
            "<ttFont>\r"  #        \r -> \n
            "  <test>\r\n"  #      \r\n -> \n
            "    a line of text\n"  #              \n
            "    escaped CR and unix newline &#13;\n"  #   &#13;\n -> \r\n
            "    escaped CR and macintosh newline &#13;\r"  #   &#13;\r -> \r\n
            "    escaped CR and windows newline &#13;\r\n"  # &#13;\r\n -> \r\n
            "  </test>\n"  #              \n
            "</ttFont>"
        )

        with BytesIO(data.encode("utf-8")) as tmp:
            reader = DebugXMLReader(tmp, TTFont())
            reader.read()
        expected = ["\n"] * 3 + ["\r", "\n"] * 3 + ["\n"]
        self.assertEqual(expected, reader.newlines)

    def test_progress(self):
        class DummyProgressPrinter(ProgressPrinter):
            def __init__(self, title, maxval=100):
                self.label = title
                self.maxval = maxval
                self.pos = 0

            def set(self, val, maxval=None):
                if maxval is not None:
                    self.maxval = maxval
                self.pos = val

            def increment(self, val=1):
                self.pos += val

            def setLabel(self, text):
                self.label = text

        data = (
            "<ttFont>\n"
            "  <test>\n"
            "    %s\n"
            "  </test>\n"
            "</ttFont>\n" % ("z" * 2 * BUFSIZE)
        ).encode("utf-8")

        dataSize = len(data)
        progressBar = DummyProgressPrinter("test")
        with BytesIO(data) as tmp:
            reader = XMLReader(tmp, TTFont(), progress=progressBar)
            self.assertEqual(progressBar.pos, 0)
            reader.read()
        self.assertEqual(progressBar.pos, dataSize // 100)
        self.assertEqual(progressBar.maxval, dataSize // 100)
        self.assertTrue("test" in progressBar.label)
        with BytesIO(b"<ttFont></ttFont>") as tmp:
            reader = XMLReader(tmp, TTFont(), progress=progressBar)
            reader.read()
        # when data size is less than 100 bytes, 'maxval' is 1
        self.assertEqual(progressBar.maxval, 1)

    def test_close_file_path(self):
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            tmp.write(b"<ttFont></ttFont>")
        reader = XMLReader(tmp.name, TTFont())
        reader.read()
        # when reading from path, the file is closed automatically at the end
        self.assertTrue(reader.file.closed)
        # this does nothing
        reader.close()
        self.assertTrue(reader.file.closed)
        os.remove(tmp.name)

    def test_close_file_obj(self):
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            tmp.write(b'<ttFont>"hello"</ttFont>')
        with open(tmp.name, "rb") as f:
            reader = XMLReader(f, TTFont())
            reader.read()
            # when reading from a file or file-like object, the latter is kept open
            self.assertFalse(reader.file.closed)
        # ... until the user explicitly closes it
        reader.close()
        self.assertTrue(reader.file.closed)
        os.remove(tmp.name)

    def test_read_sub_file(self):
        # Verifies that sub-file content is able to be read to a table.
        expectedContent = "testContent"
        expectedNameID = "1"
        expectedPlatform = "3"
        expectedLangId = "0x409"

        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            subFileData = (
                '<ttFont ttLibVersion="3.15">'
                "<name>"
                '<namerecord nameID="%s" platformID="%s" platEncID="1" langID="%s">'
                "%s"
                "</namerecord>"
                "</name>"
                "</ttFont>"
            ) % (expectedNameID, expectedPlatform, expectedLangId, expectedContent)
            tmp.write(subFileData.encode("utf-8"))

        with tempfile.NamedTemporaryFile(delete=False) as tmp2:
            fileData = (
                '<ttFont ttLibVersion="3.15">'
                "<name>"
                '<namerecord src="%s"/>'
                "</name>"
                "</ttFont>"
            ) % tmp.name
            tmp2.write(fileData.encode("utf-8"))

        ttf = TTFont()
        with open(tmp2.name, "rb") as f:
            reader = XMLReader(f, ttf)
            reader.read()
            reader.close()
            nameTable = ttf["name"]
            self.assertTrue(int(expectedNameID) == nameTable.names[0].nameID)
            self.assertTrue(int(expectedLangId, 16) == nameTable.names[0].langID)
            self.assertTrue(int(expectedPlatform) == nameTable.names[0].platformID)
            self.assertEqual(
                expectedContent,
                nameTable.names[0].string.decode(nameTable.names[0].getEncoding()),
            )

        os.remove(tmp.name)
        os.remove(tmp2.name)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from itertools import *

# Python 3.12:
if "batched" not in globals():
    # https://docs.python.org/3/library/itertools.html#itertools.batched
    def batched(iterable, n):
        # batched('ABCDEFG', 3) --> ABC DEF G
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(islice(it, n)):
            yield batch
